"""
This module encapsulates the interface to OpenFST. This
module depends on OpenFST. To enable Python support in OpenFST, use a
recent version (>=1.5.2) and compile with ``--enable_python``.
Further information can be found here:

http://www.openfst.org/twiki/bin/view/FST/PythonExtension

"""
#!/usr/bin/python
from operator import attrgetter
from itertools import product
import copy
from alphabet import createalphabet
from collections import defaultdict
EPSILON = 0xffff
import pywrapfst as fst

def TropicalWeight(param):
    """
    Returns the emulated fst TropicalWeight
    Args:
        param (str): The input
    Returns:
        bool: The arc weight
    """
    if param == (float('inf')):
        return False
    else:
        return True


class DFAState(object):
    """The DFA statess"""

    def __init__(self, cur_fst, cur_node):
        """
        Initialization function
        Args:
            sid (int): The state identifier
        Returns:
            None
        """
        self.cur_node = cur_node
        self.cur_fst = cur_fst


    def __iter__(self):
        """Iterator"""
        return iter(self.cur_fst.arcs(self.cur_node))

    def __getattribute__(self, x):
        if x == 'arcs':
            return self.cur_fst.arcs(self.cur_node)
        if x == 'stateid':
            return self.cur_node
        if x == 'initial':
            return self.cur_fst.start == self.cur_node
        if x == 'final':
            return self.cur_fst.final(self.cur_node) == fst.Weight.One(self.cur_fst.weight_type())
        return object.__getattribute__(self, x)



    def __setattr__(self, x, value):
        if x == 'arcs':
            print  'Setting arcs is not implemented'
            exit(1)
        if x == 'stateid':
            print  'Setting statedid is not implemented'
            exit(1)
        if x == 'initial':
            if value == True:
                self.cur_fst.set_start(self.cur_node)
            else:
                print  'Unsetting start is not implemented'
                exit(1)
        if x == 'final':
            if value == True:
                self.cur_fst.set_final(self.cur_node)
            else:
                self.cur_fst.set_final(self.cur_node, fst.Weight.Zero(self.cur_fst.weight_type()))
        return object.__setattr__(self, x, value)



class syms:
    """The DFA accepted symbols"""

    def __init__(self):
        """Initialize symbols"""
        self.symbols = {}
        self.reversesymbols = {}

    def __getitem__(self, char):
        """
        Finds a symbol identifier based on the input character
        Args:
            char (str): The symbol character
        Returns:
            int: The retrieved symbol identifier
        """
        return self.reversesymbols[char]

    def __setitem__(self, char, num):
        """
        Sets a symbol
        Args:
            char (str): The symbol character
            num (int): The  symbol identifier
        Returns:
           None
        """
        self.symbols[num] = char
        self.reversesymbols[char] = num

    def find(self, num):
        """
        Finds a symbol based on its identifier
        Args:
            num (int): The symbol identifier
        Returns:
            str: The retrieved symbol
        """
        return self.symbols[num]

    def items(self):
        """Returns all stored symbols
        Args:
            None
        Returns:
            dict:The included symbols
        """
        return self.symbols


class PywrapfstDFA(object):
    """A DFA implementation that uses the
    same interface with python symautomata"""
    def __init__(self, alphabet=createalphabet()):
        """
        Args:
            alphabet (list): The imput alphabet
        Returns:
            None
        """
        self.automaton = fst.Fst()
        id = self.add_state()
        self.automaton.set_start(id)
        self.alphabet = alphabet
        num = 1
        self.isyms = syms()
        self.osyms = syms()
        for char in alphabet:
            self.isyms.__setitem__(char, num)
            self.osyms.__setitem__(char, num)
            num = num + 1


    def __str__(self):
        """Describes DFA object"""
        return "This is a pywrapfstDFA object with " + `len(self.states)` + " states"

    def __getattribute__(self, x):
        if x == 'states':
            return [DFAState(self.automaton, state) for state in self.automaton.states()]
        return object.__getattribute__(self, x)



    def __getitem__(self, id):
        """
        Retrieves a state
        Args:
            state (int): State identifier
        Returns:
            DFA state: The selected dfa state
        """
        allocatedstate = DFAState(self.automaton, id)
        return allocatedstate


    def add_state(self):
        """Adds a new state"""
        return self.automaton.add_state()

    def add_arc(self, src, dst, char):
        """Adds a new Arc
        Args:
            src (int): The source state identifier
            dst (int): The destination state identifier
            char (str): The character for the transition
        Returns:
            None
        """
        if src not in self.automaton.states():
            self.add_state()
        arc = fst.Arc(self.isyms[char], self.osyms[char],  fst.Weight.One(self.automaton.weight_type()), dst)
        self.automaton.add_arc(src, arc)



    def fixminimized(self, alphabet):
        """
        After pyfst minimization,
        all unused arcs are removed,
        and all sink states are removed.
        However this may break compatibility.
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        endstate = self.add_state()
        for state in self.states:
            for char in alphabet:
                found = 0
                for arc in state.arcs:
                    if self.isyms.find(arc.ilabel) == char:
                        found = 1
                        break
                if found == 0:
                    self.add_arc(state.stateid, endstate, char)

        self[endstate].final = False

        for char in alphabet:
            self.add_arc(endstate, endstate, char)

    def _addsink(self, alphabet):
        """
        Adds a sink state
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        endstate = self.add_state()
        for state in self.states:
            for char in alphabet:
                found = 0
                for arc in state.arcs:
                    if self.isyms.find(arc.ilabel) == char:
                        found = 1
                        break
                if found == 0:
                    self.add_arc(state.stateid, endstate, char)

        self[endstate].final = False

        for char in alphabet:
            self.add_arc(endstate, endstate, char)
        self.determinize()

    def consume_input(self, inp):
        """
        Return True/False if the machine accepts/reject the input.
        Args:
            inp (str): input string to be consumed
        Returns:
            bool: A true or false value depending on if the DFA
                accepts the provided input
        """
        cur_state = sorted(
            self.states,
            key=attrgetter('initial'),
            reverse=True)[0]
        while len(inp) > 0:
            found = False
            for arc in cur_state.arcs:
                if self.isyms.find(arc.ilabel) == inp[0]:
                    cur_state = self[arc.nextstate]
                    inp = inp[1:]
                    found = True
                    break
            if not found:
                return False
        return cur_state.final

    def empty(self):
        """
        Return True if the DFA accepts the empty language.
        """
        # self.minimize()
        return len(list(self.states)) == 0

    def complement(self, alphabet):
        """
        Returns the complement of DFA
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        self._addsink(alphabet)
        for state in self.automaton.states():
            if self.automaton.final(state) == fst.Weight.One(self.automaton.weight_type()):
                self.automaton.set_final(state, fst.Weight.Zero(self.automaton.weight_type()))
            else:
                self.automaton.set_final(state, fst.Weight.One(self.automaton.weight_type()))


    def init_from_acceptor(self, acceptor):
        """
        Adds a sink state
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        self.automaton = acceptor.automaton

    def save(self, txt_fst_file_name):
        """
        Save the machine in the openFST format in the file denoted by
        txt_fst_file_name.
        Args:
            txt_fst_file_name (str): The output file
        Returns:
            None
        """
        output_filename = open(txt_fst_file_name, 'w+')
        states = sorted(self.states, key=attrgetter('initial'), reverse=True)
        for state in states:
            for arc in state.arcs:
                if arc.ilabel == 0:
                    continue
                itext = self.isyms.find(arc.ilabel)
                otext = self.osyms.find(arc.ilabel)
                output_filename.write(
                    '{}\t{}\t{}\t{}\n'.format(
                        state.stateid,
                        arc.nextstate,
                        itext.encode('hex'),
                        otext.encode('hex')))
            if state.final:
                output_filename.write('{}\n'.format(state.stateid))
        output_filename.close()

    def load(self, txt_fst_file_name):
        """
        Save the transducer in the text file format of OpenFST.
        The format is specified as follows:
            arc format: src dest ilabel olabel [weight]
            final state format: state [weight]
        lines may occur in any order except initial state must be first line
        Args:
            txt_fst_file_name (str): The input file
        Returns:
            None
        """
        with open(txt_fst_file_name, 'r') as input_filename:
            for line in input_filename:
                line = line.strip()
                split_line = line.split()
                if len(split_line) == 1:
                    self[int(split_line[0])].final = True
                else:
                    if int(split_line[1]) == 0:
                        continue
                    self.add_arc(int(split_line[0]), int(split_line[1]),
                                 split_line[2].decode('hex'))


    def minimize(self):
        """Minimizes the DFA using Hopcroft algorithm"""
        self.determinize()
        self.automaton.minimize()

    def intersect(self, other):
        """Constructs an unminimized DFA recognizing
        the intersection of the languages of two given DFAs.
        Args:
            other (DFA): The other DFA that will be used
                         for the intersect operation
        Returns:
        Returns:
            DFA: The resulting DFA
        """
        self.automaton = fst.intersect(self.automaton, other.automaton)
        return  self

    def __and__(self, other):
        """Constructs an unminimized DFA recognizing
               the intersection of the languages of two given DFAs.
               Args:
                   other (DFA): The other DFA that will be used
                                for the intersect operation
               Returns:
                   DFA: The resulting DFA
               """
        self.intersect(other)
        return self

    def symmetric_difference(self, other):
        """Constructs an unminimized DFA recognizing
        the symmetric difference of the languages of two given DFAs.
        Args:
            other (DFA): The other DFA that will be used
                         for the symmetric difference operation
        Returns:
            DFA: The resulting DFA
        """
        self.automaton.symmetric_difference(other)
        return  self

    def union(self, other):
        """Constructs an unminimized DFA recognizing the union of the languages of two given DFAs.
        Args:
            other (DFA): The other DFA that will be used
                         for the union operation
        Returns:
            DFA: The resulting DFA
        """
        self.automaton.union(other.automaton)
        return self

    def __or__(self, other):
        """Constructs an unminimized DFA recognizing the union of the languages of two given DFAs.
                Args:
                    other (DFA): The other DFA that will be used
                                 for the union operation
                Returns:
                    DFA: The resulting DFA
                """
        self.union(other)
        return self

    def determinize(self):
        """
        Transforms a Non Deterministic DFA into a Deterministic
        Args:
            None
        Returns:
            DFA: The resulting DFA
        """
        # This function is not necessary
        self.automaton = fst.determinize(self.automaton)
        return  self

    def invert(self):
        """Inverts the DFA final states"""
        self.automaton.inver()

    def difference(self, other):
        """Performs the Diff operation between two atomata
        Args:
            other (DFA): The other DFA that will be used
                         for the difference operation
        Returns:
            DFA: The resulting DFA
        """
        other.invert()
        self.intersect(other)
        return self

    def __sub__(self, other):
        """Performs the Diff operation between two atomata
        Args:
            other (DFA): The other DFA that will be used
                        for the difference operation
        Returns:
            DFA: The resulting DFA
        """
        self.difference(other)
        return self





"""This module contains the PDA implementation"""
# !/usr/bin/python

try:
    print 'Checking for pythonpda module:',
    print 'OK'
    from pythonpda import PythonPDA, PDAState

    class PDA(PythonPDA):
        """This is the structure for a PDA"""

        def shortest_string(self):
            """
            Uses BFS in order to find the shortest string
            Args:
                None
            Returns:
                str: The shortest string
            """
            from pdadiff import PdaDiff
            ops = PdaDiff(None, None, self.alphabet)
            ops.mmc = self
            return ops.get_string()

        def diff(self, mmb):
            """
            Automata Diff operation
            """
            from pdadiff import PdaDiff
            ops = PdaDiff(self, mmb, self.alphabet)
            mmc = ops.diff()
            return mmc

        def  consume_input(self, str):
            """
            Not implemented
            """
            return 0xffff

except ImportError:
    print 'FAIL'



"""
This module transforms a pyfst DFA to regular expressions
using the Brzozowski Algebraic Method
"""
import sys
from collections import OrderedDict
from operator import attrgetter

from alphabet import createalphabet
from dfa import DFA
from flex2fst import Flexparser


class Brzozowski:
    """
    This class transforms a pyfst DFA to regular expressions
    using the Brzozowski Algebraic Method
    """

    def __init__(self, input_fst_a, alphabet=None):
        """
        Args:
            input_fst_a (DFA): The DFA states
            alphabet (list): the input alphabet
        Returns:
            None
        """
        if alphabet is None:
            alphabet = createalphabet()
        self.alphabet = alphabet
        self.mma = DFA(self.alphabet)
        self.mma.init_from_acceptor(input_fst_a)
        # a is a matrix that holds all transitions
        # If there is a transition from state 0 to state 1 with the symbol x
        # then a[0][1]=x
        self.A = {}
        # B[n] holds the regular expression that describes how a final state
        # can be reached from state n
        self.B = {}
        self.epsilon = ''
        self.empty = None

    def _bfs_sort(self, start):
        """
        maintain a map of states distance using BFS
        Args:
            start (fst state): The initial DFA state
        Returns:
            list: An ordered list of DFA states
                  using path distance
        """
        pathstates = {}
        # maintain a queue of nodes to be visited. Both current and previous
        # node must be included.
        queue = []
        # push the first path into the queue
        queue.append([0, start])
        pathstates[start.stateid] = 0
        while queue:
            # get the first node from the queue
            leaf = queue.pop(0)
            node = leaf[1]
            pathlen = leaf[0]
            # enumerate all adjacent nodes, construct a new path and push it
            # into the queue
            for arc in node.arcs:
                next_state = self.mma[arc.nextstate]
                if next_state.stateid not in pathstates:
                    queue.append([pathlen + 1, next_state])
                    pathstates[next_state.stateid] = pathlen + 1
        orderedstatesdict = OrderedDict(
            sorted(
                pathstates.items(),
                key=lambda x: x[1],
                reverse=False))
        for state in self.mma.states:
            orderedstatesdict[state.stateid] = state
        orderedstates = [x[1] for x in list(orderedstatesdict.items())]
        return orderedstates

    def star(self, input_string):
        """
        Kleene star operation
        Args:
            input_string (str): The string that the kleene star will be made
        Returns:
            str: The applied Kleene star operation on the input string
        """
        if input_string != self.epsilon and input_string != self.empty:
            return "(" + input_string + ")*"
        else:
            return ""

    def _brzozowski_algebraic_method_init(self):
        """Initialize Brzozowski Algebraic Method"""
        # Initialize B
        for state_a in self.mma.states:
            if state_a.final:
                self.B[state_a.stateid] = self.epsilon
            else:
                self.B[state_a.stateid] = self.empty
            # Initialize A
            for state_b in self.mma.states:
                self.A[state_a.stateid, state_b.stateid] = self.empty
                for arc in state_a.arcs:
                    if arc.nextstate == state_b.stateid:
                        self.A[state_a.stateid, state_b.stateid] = \
                            self.mma.isyms.find(arc.ilabel)

    def _brzozowski_algebraic_method_solve(self):
        """Perform Brzozowski Algebraic Method"""
        orderedstates = self._bfs_sort(
            sorted(
                self.mma.states,
                key=attrgetter('initial'),
                reverse=True)[0])
        for n in range(len(orderedstates) - 1, 0, -1):
            # print "n:" + repr(n)
            if self.A[
                    orderedstates[n].stateid,
                    orderedstates[n].stateid] != self.empty:
                # B[n] := star(A[n,n]) . B[n]
                if self.B[orderedstates[n].stateid] != self.empty:
                    self.B[orderedstates[n].stateid] = \
                        self.star(self.A[orderedstates[n].stateid, orderedstates[n].stateid]) \
                        + self.B[orderedstates[n].stateid]
                else:
                    self.B[orderedstates[n].stateid] = self.star(
                        self.A[orderedstates[n].stateid, orderedstates[n].stateid])
                for j in range(0, n):
                    # A[n,j] := star(A[n,n]) . A[n,j]
                    if self.A[
                            orderedstates[n].stateid,
                            orderedstates[j].stateid] != self.empty:
                        self.A[
                            orderedstates[n].stateid,
                            orderedstates[j].stateid] = \
                            self.star(self.A[orderedstates[n].stateid, orderedstates[n].stateid]) \
                            + self.A[orderedstates[n].stateid, orderedstates[j].stateid]
                    else:
                        self.A[orderedstates[n].stateid, orderedstates[j].stateid] = self.star(
                            self.A[orderedstates[n].stateid, orderedstates[n].stateid])
            for i in range(0, n):
                # B[i] += A[i,n] . B[n]
                newnode = None
                if self.A[orderedstates[i].stateid, orderedstates[n].stateid] != self.empty \
                        and self.B[orderedstates[n].stateid] != self.empty:
                    newnode = self.A[orderedstates[i].stateid, orderedstates[
                        n].stateid] + self.B[orderedstates[n].stateid]
                elif self.A[orderedstates[i].stateid, orderedstates[n].stateid] != self.empty:
                    newnode = self.A[
                        orderedstates[i].stateid,
                        orderedstates[n].stateid]
                elif self.B[orderedstates[n].stateid] != self.empty:
                    newnode = self.B[orderedstates[n].stateid]
                if self.B[orderedstates[i].stateid] != self.empty:
                    if newnode is not None:
                        self.B[orderedstates[i].stateid] += newnode
                else:
                    self.B[orderedstates[i].stateid] = newnode
                for j in range(0, n):
                    # A[i,j] += A[i,n] . A[n,j]
                    newnode = None
                    if self.A[
                            orderedstates[i].stateid,
                            orderedstates[n].stateid] != self.empty \
                            and self.A[orderedstates[n].stateid, orderedstates[j].stateid] \
                                    != self.empty:
                        newnode = self.A[orderedstates[i].stateid, orderedstates[
                            n].stateid] + self.A[orderedstates[n].stateid, orderedstates[j].stateid]
                    elif self.A[orderedstates[i].stateid, orderedstates[n].stateid] != self.empty:
                        newnode = self.A[
                            orderedstates[i].stateid,
                            orderedstates[n].stateid]
                    elif self.A[orderedstates[n].stateid, orderedstates[j].stateid] != self.empty:
                        newnode = self.A[
                            orderedstates[n].stateid,
                            orderedstates[j].stateid]
                    if self.A[
                            orderedstates[i].stateid,
                            orderedstates[j].stateid] != self.empty:
                        if newnode is not None:
                            self.A[
                                orderedstates[i].stateid,
                                orderedstates[j].stateid] += newnode
                    else:
                        self.A[
                            orderedstates[i].stateid,
                            orderedstates[j].stateid] = newnode


    def get_regex(self):
        """Generate regular expressions from DFA automaton"""
        self._brzozowski_algebraic_method_init()
        self._brzozowski_algebraic_method_solve()
        return self.B


def main():
    """
    Testing function for DFA brzozowski algebraic method Operation
    """
    argv = sys.argv
    if len(argv) < 2:
        targetfile = 'target.y'
    else:
        targetfile = argv[1]
    print 'Parsing ruleset: ' + targetfile,
    flex_a = Flexparser()
    mma = flex_a.yyparse(targetfile)
    print 'OK'
    print 'Perform minimization on initial automaton:',
    mma.minimize()
    print 'OK'
    print 'Perform Brzozowski on minimal automaton:',
    brzozowski_a = Brzozowski(mma)
    mma_regex = brzozowski_a.get_regex()
    print mma_regex

if __name__ == '__main__':
    main()



"""
This module transforms a pyfst DFA to regular expressions
using the State Removal Method
"""
from sys import argv
from operator import attrgetter
from alphabet import createalphabet
from dfa import DFA
from flex2fst import Flexparser


class StateRemoval:
    """Transforms a pyfst DFA to regular expressions"""
    def __init__(self, input_fst_a, alphabet=None):
        """
        Args:
            input_fst_a (DFA): The DFA states
            alphabet (list): the input alphabet
        Returns:
            None
        """
        if alphabet is None:
            alphabet = createalphabet()
        self.mma = DFA(self.alphabet)
        self.mma.init_from_acceptor(input_fst_a)
        self.alphabet = alphabet

        self.l_transitions = {}
        self.epsilon = ''
        self.empty = None

    def star(self, input_string):
        """
        Kleene star operation
        Args:
            input_string (str): The string that the kleene star will be made
        Returns:
            str: The applied Kleene star operation on the input string
        """
        if input_string != self.epsilon and input_string != self.empty:
            return "(" + input_string + ")*"
        else:
            return ""

    def _state_removal_init(self):
        """State Removal Operation Initialization"""
        # First, we remove all multi-edges:
        for state_i in self.mma.states:
            for state_j in self.mma.states:
                if state_i.stateid == state_j.stateid:
                    self.l_transitions[
                        state_i.stateid, state_j.stateid] = self.epsilon
                else:
                    self.l_transitions[
                        state_i.stateid, state_j.stateid] = self.empty

                for arc in state_i.arcs:
                    if arc.nextstate == state_j.stateid:
                        if self.l_transitions[state_i.stateid, state_j.stateid] != self.empty:
                            self.l_transitions[state_i.stateid, state_j.stateid] \
                                += self.mma.isyms.find(arc.ilabel)
                        else:
                            self.l_transitions[state_i.stateid, state_j.stateid] = \
                                self.mma.isyms.find(arc.ilabel)

    def _state_removal_remove(self, k):
        """
        State Removal Remove operation
        l_transitions[i,i] += l_transitions[i,k] . star(l_transitions[k,k]) . l_transitions[k,i]
        l_transitions[j,j] += l_transitions[j,k] . star(l_transitions[k,k]) . l_transitions[k,j]
        l_transitions[i,j] += l_transitions[i,k] . star(l_transitions[k,k]) . l_transitions[k,j]
        l_transitions[j,i] += l_transitions[j,k] . star(l_transitions[k,k]) . l_transitions[k,i]
        Args:
            k (int): The node that will be removed
        Returns:
            None
        """
        for state_i in self.mma.states:
            for state_j in self.mma.states:
                if self.l_transitions[state_i.stateid, k] != self.empty:
                    l_ik = self.l_transitions[state_i.stateid, k]
                else:
                    l_ik = ""
                if self.l_transitions[state_j.stateid, k] != self.empty:
                    l_jk = self.l_transitions[state_j.stateid, k]
                else:
                    l_jk = ""
                if self.l_transitions[k, state_i.stateid] != self.empty:
                    l_ki = self.l_transitions[k, state_i.stateid]
                else:
                    l_ki = ""
                if self.l_transitions[k, state_j.stateid] != self.empty:
                    l_kj = self.l_transitions[k, state_j.stateid]
                else:
                    l_kj = ""

                if self.l_transitions[state_i.stateid, state_i.stateid] != self.empty:
                    self.l_transitions[state_i.stateid, state_i.stateid] += l_ik + \
                        self.star(self.l_transitions[k, k]) + l_ki
                else:
                    self.l_transitions[state_i.stateid, state_i.stateid] = l_ik + \
                        self.star(self.l_transitions[k, k]) + l_ki

                if self.l_transitions[state_j.stateid, state_j.stateid] != self.empty:
                    self.l_transitions[state_j.stateid, state_j.stateid] += l_jk + \
                        self.star(self.l_transitions[k, k]) + l_kj
                else:
                    self.l_transitions[state_j.stateid, state_j.stateid] = l_jk + \
                        self.star(self.l_transitions[k, k]) + l_kj

                if self.l_transitions[state_i.stateid, state_j.stateid] != self.empty:
                    self.l_transitions[state_i.stateid, state_j.stateid] += l_ik + \
                        self.star(self.l_transitions[k, k]) + l_kj
                else:
                    self.l_transitions[state_i.stateid, state_j.stateid] = l_ik + \
                        self.star(self.l_transitions[k, k]) + l_kj

                if self.l_transitions[state_j.stateid, state_i.stateid] != self.empty:
                    self.l_transitions[state_j.stateid, state_i.stateid] += l_jk + \
                        self.star(self.l_transitions[k, k]) + l_ki
                else:
                    self.l_transitions[state_j.stateid, state_i.stateid] = l_jk + \
                        self.star(self.l_transitions[k, k]) + l_ki

    def _state_removal_solve(self):
        """The State Removal Operation"""
        initial = sorted(
            self.mma.states,
            key=attrgetter('initial'),
            reverse=True)[0].stateid
        for state_k in self.mma.states:
            if state_k.final:
                continue
            if state_k.stateid == initial:
                continue
            self._state_removal_remove(state_k.stateid)

        print self.l_transitions
        return self.l_transitions

    def get_regex(self):
        """Regular Expression Generation"""
        self._state_removal_init()
        self._state_removal_solve()
        return self.l_transitions[0]


def main():
    """Testing function for DFA _Brzozowski Operation"""
    if len(argv) < 2:
        targetfile = 'target.y'
    else:
        targetfile = argv[1]
    print 'Parsing ruleset: ' + targetfile,
    flex_a = Flexparser()
    mma = flex_a.yyparse(targetfile)
    print 'OK'
    print 'Perform minimization on initial automaton:',
    mma.minimize()
    print 'OK'
    print 'Perform StateRemoval on minimal automaton:',
    state_removal = StateRemoval(mma)
    mma_regex = state_removal.get_regex()
    print mma_regex

if __name__ == '__main__':
    main()


"""
This module retrieves a simple string from a PDA
using the state removal method
"""
from pda import PDAState


class PdaString():
    """Retrieves a string from a PDA"""
    statediag = []
    quickresponse = {}
    quickresponse_types = {}

    def __init__(self):
        """Class Initialization"""
        pass
    
    def _combine_rest_push(self):
        """Combining Rest and Push States"""
        new = []
        change = 0
        # DEBUG
        # logging.debug('Combining Rest and Push')
        i = 0
        examinetypes = self.quickresponse_types[3]
        for state in examinetypes:
            if state.type == 3:
                for nextstate_id in state.trans.keys():
                    found = 0
                    # if nextstate_id != state.id:
                    if nextstate_id in self.quickresponse:
                        examines = self.quickresponse[nextstate_id]
                        for examine in examines:
                            if examine.id == nextstate_id and examine.type == 1:
                                temp = PDAState()
                                temp.type = 1
                                temp.sym = examine.sym
                                temp.id = state.id
                                for nextnextstate_id in examine.trans:
                                    # if nextnextstate_id != examine.id :

                                    for x_char in state.trans[nextstate_id]:
                                        for z_char in examine.trans[
                                                nextnextstate_id]:
                                            if nextnextstate_id not in temp.trans:
                                                temp.trans[
                                                    nextnextstate_id] = []
                                            if x_char != 0 and z_char != 0:
                                                temp.trans[
                                                    nextnextstate_id].append(x_char + z_char)
                                                # DEBUGprint 'transition is now
                                                # '+x_char +' + '+ z_char
                                            elif x_char != 0 and z_char == 0:
                                                temp.trans[
                                                    nextnextstate_id].append(x_char)
                                                # DEBUGprint 'transition is now
                                                # '+x_char
                                            elif x_char == 0 and z_char != 0:
                                                temp.trans[
                                                    nextnextstate_id].append(z_char)
                                                # DEBUGprint 'transition is now
                                                # '+z_char
                                            elif x_char == 0 and z_char == 0:
                                                temp.trans[
                                                    nextnextstate_id].append(0)
                                                # DEBUGprint 'transition is now
                                                # empty'
                                            else:
                                                pass
                                found = 1
                                new.append(temp)

                    if found == 1:
                        # print 'Lets combine one with id '+`state.id`+'(rest)
                        # and one with id '+`nextstate_id`+'(push)'
                        change = 1
                        # del(state.trans[nextstate_id])
            i = i + 1
        if change == 0:
            return []
        else:
            return new

    def _combine_push_rest(self):
        """Combining Push and Rest"""
        new = []
        change = 0
        # DEBUG
        # logging.debug('Combining Push and Rest')
        i = 0
        examinetypes = self.quickresponse_types[1]
        for state in examinetypes:
            if state.type == 1:
                for nextstate_id in state.trans.keys():
                    found = 0
                    # if nextstate_id != state.id:
                    if nextstate_id in self.quickresponse:
                        examines = self.quickresponse[nextstate_id]
                        for examine in examines:
                            if examine.id == nextstate_id and examine.type == 3:
                                temp = PDAState()
                                temp.type = 1
                                temp.sym = state.sym
                                temp.id = state.id
                                for nextnextstate_id in examine.trans:
                                    # if nextnextstate_id != examine.id :
                                    for x_char in state.trans[nextstate_id]:
                                        for z_char in examine.trans[
                                                nextnextstate_id]:
                                            if nextnextstate_id not in temp.trans:
                                                temp.trans[
                                                    nextnextstate_id] = []
                                            if x_char != 0 and z_char != 0:
                                                temp.trans[
                                                    nextnextstate_id].append(x_char + z_char)
                                                # DEBUGprint 'transition is now
                                                # '+x_char +' + '+ z_char
                                            elif x_char != 0 and z_char == 0:
                                                temp.trans[
                                                    nextnextstate_id].append(x_char)
                                                # DEBUGprint 'transition is now
                                                # '+x_char
                                            elif x_char == 0 and z_char != 0:
                                                temp.trans[
                                                    nextnextstate_id].append(z_char)
                                                # DEBUGprint 'transition is now
                                                # '+z_char
                                            elif x_char == 0 and z_char == 0:
                                                temp.trans[
                                                    nextnextstate_id].append(0)
                                                # DEBUGprint 'transition is now
                                                # empty'
                                            else:
                                                pass
                                found = 1
                                new.append(temp)

                    if found == 1:
                        # DEBUGprint 'Lets combine one with id
                        # '+`state.id`+'(push) and one with id
                        # '+`nextstate_id`+'(rest)'
                        change = 1
                        del state.trans[nextstate_id]
            i = i + 1
        if change == 0:
            return []
        else:
            return new


    def _combine_pop_rest(self):
        """Combining Pop and Rest"""
        new = []
        change = 0
        # DEBUG
        # logging.debug('Combining Pop and Rest')
        i = 0
        examinetypes = self.quickresponse_types[2]
        for state in examinetypes:
            if state.type == 2:
                for nextstate_id in state.trans.keys():
                    found = 0
                    # if nextstate_id != state.id:
                    if nextstate_id in self.quickresponse:
                        examines = self.quickresponse[nextstate_id]
                        for examine in examines:
                            if examine.id == nextstate_id and examine.type == 3:
                                if state.sym != 0:
                                    temp = PDAState()
                                    temp.type = 2
                                    temp.sym = state.sym
                                    temp.id = state.id
                                    for nextnextstate_id in examine.trans:
                                        # if nextnextstate_id != examine.id:

                                        for x_char in state.trans[nextstate_id]:
                                            for z_char in examine.trans[
                                                    nextnextstate_id]:
                                                if nextnextstate_id not in temp.trans:
                                                    temp.trans[
                                                        nextnextstate_id] = []
                                                if x_char != 0 and z_char != 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(x_char + z_char)
                                                    # DEBUGprint 'transition is
                                                    # now '+x_char +' + '+ z_char
                                                elif x_char != 0 and z_char == 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(x_char)
                                                    # DEBUGprint 'transition is
                                                    # now '+x_char
                                                elif x_char == 0 and z_char != 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(z_char)
                                                    # DEBUGprint 'transition is
                                                    # now '+z_char
                                                elif x_char == 0 and z_char == 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(0)
                                                    # DEBUGprint 'transition is
                                                    # now empty'
                                                else:
                                                    pass
                                    found = 1
                                    new.append(temp)
                                else:
                                    for nextnextstate_id in examine.trans:
                                        # if nextnextstate_id != examine.id:

                                        for x_char in state.trans[nextstate_id]:
                                            temp = PDAState()
                                            temp.type = 2
                                            temp.id = state.id
                                            temp.sym = x_char
                                            temp.trans[nextnextstate_id] = []
                                            for z_char in examine.trans[
                                                    nextnextstate_id]:
                                                if z_char != 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(z_char)
                                                    # DEBUGprint 'transition is
                                                    # now '+z_char
                                                elif z_char == 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(0)
                                                    # DEBUGprint 'transition is
                                                    # now empty'
                                                else:
                                                    pass
                                            found = 1
                                            new.append(temp)

                    if found == 1:
                        # DEBUGprint 'Lets combine one with id
                        # '+`state.id`+'(push) and one with id
                        # '+`nextstate_id`+'(rest)'
                        change = 1
                        del state.trans[nextstate_id]
            i = i + 1
        if change == 0:
            return []
        else:
            return new

    def _combine_rest_rest(self):
        """Combining Rest and Rest"""
        new = []
        change = 0
        # DEBUG
        # logging.debug('Combining Rest and Rest')
        i = 0
        examinetypes = self.quickresponse_types[3]
        for state in examinetypes:
            if state.type == 3:
                found = 0
                for nextstate_id in state.trans.keys():
                    secondfound = 0
                    # if nextstate_id != state.id:
                    if nextstate_id in self.quickresponse:
                        examines = self.quickresponse[nextstate_id]
                        for examine in examines:
                            if examine.id == nextstate_id and examine.type == 3:
                                temp = PDAState()
                                temp.type = 3
                                temp.sym = state.sym
                                temp.id = state.id
                                for nextnextstate_id in examine.trans:
                                    if nextnextstate_id != examine.id:
                                        for x_char in state.trans[nextstate_id]:
                                            for z_char in examine.trans[
                                                    nextnextstate_id]:
                                                if nextnextstate_id not in temp.trans:
                                                    temp.trans[
                                                        nextnextstate_id] = []
                                                if x_char != 0 and z_char != 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(x_char + z_char)
                                                    # DEBUGprint 'transition is
                                                    # now '+x_char +' + '+ z_char
                                                elif x_char != 0 and z_char == 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(x_char)
                                                    # DEBUGprint 'transition is
                                                    # now '+x_char
                                                elif x_char == 0 and z_char != 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(z_char)
                                                    # DEBUGprint 'transition is
                                                    # now '+z_char
                                                elif x_char == 0 and z_char == 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(0)
                                                    # DEBUGprint 'transition is
                                                    # now empty'
                                                else:
                                                    pass
                                                secondfound = 1
                                if secondfound == 1:
                                    new.append(temp)
                                    found = 1

                    if found == 1:
                        # DEBUGprint 'Lets combine one with id
                        # '+`state.id`+'(rest) and one with id
                        # '+`nextstate_id`+'(rest)'
                        change = 1
                        del state.trans[nextstate_id]
            i = i + 1
        if change == 0:
            return []
        else:
            return new

    def _combine_push_pop(self):
        """Combining Push and Pop"""
        new = []
        change = 0
        # DEBUG
        # logging.debug('Combining Push and Pop')
        i = 0
        examinetypes = self.quickresponse_types[1]
        for state in examinetypes:
            if state.type == 1:
                found = 0
                for nextstate_id in state.trans.keys():
                    # if nextstate_id != state.id:
                    if nextstate_id in self.quickresponse:
                        examines = self.quickresponse[nextstate_id]
                        for examine in examines:
                            secondfound = 0
                            if examine.id == nextstate_id and examine.type == 2:
                                temp = PDAState()
                                temp.type = 3
                                temp.sym = 0
                                temp.id = state.id
                                if examine.sym == 0:
                                    for nextnextstate_id in examine.trans:
                                        # if nextnextstate_id != examine.id :
                                        for z_char in examine.trans[
                                                nextnextstate_id]:
                                            if state.sym == z_char:
                                                for x_char in state.trans[
                                                        nextstate_id]:
                                                    # DEBUGprint state.sym+' vs
                                                    # '+z_char
                                                    if nextnextstate_id not in temp.trans:
                                                        temp.trans[
                                                            nextnextstate_id] = []
                                                    if x_char != 0:
                                                        temp.trans[
                                                            nextnextstate_id].append(x_char)
                                                        # DEBUGprint
                                                        # 'transition is now
                                                        # '+x_char
                                                    else:
                                                        temp.trans[
                                                            nextnextstate_id].append(0)
                                                        # DEBUGprint
                                                        # 'transition is now
                                                        # empty'

                                                    secondfound = 1

                                elif state.sym == examine.sym:
                                    for nextnextstate_id in examine.trans:
                                        # if nextnextstate_id != examine.id :
                                        for x_char in state.trans[nextstate_id]:
                                            for z_char in examine.trans[
                                                    nextnextstate_id]:
                                                if nextnextstate_id not in temp.trans:
                                                    temp.trans[
                                                        nextnextstate_id] = []
                                                if x_char != 0 and z_char != 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(x_char + z_char)
                                                    # DEBUGprint 'transition is
                                                    # now '+x_char +' + '+ z_char
                                                elif x_char != 0 and z_char == 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(x_char)
                                                    # DEBUGprint 'transition is
                                                    # now '+x_char
                                                elif x_char == 0 and z_char != 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(z_char)
                                                    # DEBUGprint 'transition is
                                                    # now '+z_char
                                                elif x_char == 0 and z_char == 0:
                                                    temp.trans[
                                                        nextnextstate_id].append(0)
                                                    # DEBUGprint 'transition is
                                                    # now empty'
                                                else:
                                                    pass
                                                secondfound = 1

                                if secondfound == 1:
                                    new.append(temp)
                                    found = 1
                    if found == 1:
                        # DEBUGprint 'Lets combine one with id
                        # '+`state.id`+'(push) and one with id
                        # '+`nextstate_id`+'(pop)'
                        change = 1
                        # DEBUGprint 'delete '+`nextstate_id`+' from
                        # '+`state.id`
                        del state.trans[nextstate_id]
            i = i + 1
        if change == 0:
            return []
        else:
            return new

    def _check(self, accepted):
        """_check for string existence"""
        # logging.debug('A check is now happening...')
        # for key in self.statediag[1].trans:
        #    logging.debug('transition to '+`key`+" with "+self.statediag[1].trans[key][0])
        total = []
        if 1 in self.quickresponse:
            total = total + self.quickresponse[1]
        if (1, 0) in self.quickresponse:
            total = total + self.quickresponse[(1, 0)]
        for key in total:
            if (key.id == 1 or key.id == (1, 0)) and key.type == 3:
                if accepted is None:
                    if 2 in key.trans:
                        # print 'Found'
                        return key.trans[2]
                else:
                    for state in accepted:
                        if (2, state) in key.trans:
                            # print 'Found'
                            return key.trans[(2, state)]
        return -1

    def _stage(self, accepted, count=0):
        """This is a repeated state in the state removal algorithm"""
        new5 = self._combine_rest_push()
        new1 = self._combine_push_pop()
        new2 = self._combine_push_rest()
        new3 = self._combine_pop_rest()
        new4 = self._combine_rest_rest()
        new = new1 + new2 + new3 + new4 + new5
        del new1
        del new2
        del new3
        del new4
        del new5
        if len(new) == 0:
            # self.printer()
            # print 'PDA is empty'
            # logging.debug('PDA is empty')
            return None

        self.statediag = self.statediag + new
        del new
        # print 'cleaning...'

        # It is cheaper to create a new array than to use the old one and
        # delete a key
        newstates = []
        for key in self.statediag:
            if len(key.trans) == 0 or key.trans == {}:
                # rint 'delete '+`key.id`
                # self.statediag.remove(key)
                pass
            else:
                newstates.append(key)

        del self.statediag
        self.statediag = newstates
        self.quickresponse = {}
        self.quickresponse_types = {}
        self.quickresponse_types[0] = []
        self.quickresponse_types[1] = []
        self.quickresponse_types[2] = []
        self.quickresponse_types[3] = []
        self.quickresponse_types[4] = []
        for state in self.statediag:
            if state.id not in self.quickresponse:
                self.quickresponse[state.id] = [state]
            else:
                self.quickresponse[state.id].append(state)
            self.quickresponse_types[state.type].append(state)

            # else:
            #     print `key.id`+' (type: '+`key.type`+' and sym:'+`key.sym`+')'
            #     print key.trans
        # print 'checking...'
        exists = self._check(accepted)
        if exists == -1:
            # DEBUGself.printer()
            # raw_input('next step?')
            return self._stage(accepted, count + 1)
        else:
            # DEBUGself.printer()
            # print 'Found '
            print exists
            # return self._stage(accepted, count+1)
            return exists

    def printer(self):
        """Visualizes the current state"""
        for key in self.statediag:
            if key.trans is not None and len(key.trans) > 0:
                print '****** ' + repr(key.id) + '(' + repr(key.type)\
                      + ' on sym ' + repr(key.sym) + ') ******'
                print key.trans

    def init(self, states, accepted):
        """Initialization of the indexing dictionaries"""
        for key in states:
            self.statediag.append(states[key])
        self.quickresponse = {}
        self.quickresponse_types = {}
        self.quickresponse_types[0] = []
        self.quickresponse_types[1] = []
        self.quickresponse_types[2] = []
        self.quickresponse_types[3] = []
        self.quickresponse_types[4] = []
        for state in self.statediag:
            if state.id not in self.quickresponse:
                self.quickresponse[state.id] = [state]
            else:
                self.quickresponse[state.id].append(state)
            self.quickresponse_types[state.type].append(state)
        # self.printer()
        # raw_input('next stepA?')
        return self._stage(accepted, 0)



"""
This module performs all basic DFA operations, without pyfst
"""
#!/usr/bin/python
from operator import attrgetter
from itertools import product
import copy
from alphabet import createalphabet
from collections import defaultdict
EPSILON = 0xffff

def TropicalWeight(param):
    """
    Returns the emulated fst TropicalWeight
    Args:
        param (str): The input
    Returns:
        bool: The arc weight
    """
    if param == (float('inf')):
        return False
    else:
        return True


class DFAState:
    """The DFA statess"""

    def __init__(self, sid=None):
        """
        Initialization function
        Args:
            sid (int): The state identifier
        Returns:
            None
        """
        self.final = False
        self.initial = False
        self.stateid = sid
        self.arcs = []

    def __iter__(self):
        """Iterator"""
        return iter(self.arcs)



class DFAArc:
    """The DFA transition"""

    def __init__(self, srcstate_id, nextstate_id, ilabel=None):
        """
        The initialization function
        Args:
            srcstate_id (int): The source state identifier
            nextstate_id (int): The destination state identifier
            ilabel (str): The symbol corresponding to character for the transition
        """
        self.srcstate = srcstate_id
        self.nextstate = nextstate_id
        self.ilabel = ilabel


class syms:
    """The DFA accepted symbols"""

    def __init__(self):
        """Initialize symbols"""
        self.symbols = {}
        self.reversesymbols = {}

    def __getitem__(self, char):
        """
        Finds a symbol identifier based on the input character
        Args:
            char (str): The symbol character
        Returns:
            int: The retrieved symbol identifier
        """
        return self.reversesymbols[char]

    def __setitem__(self, char, num):
        """
        Sets a symbol
        Args:
            char (str): The symbol character
            num (int): The  symbol identifier
        Returns:
           None
        """
        self.symbols[num] = char
        self.reversesymbols[char] = num

    def find(self, num):
        """
        Finds a symbol based on its identifier
        Args:
            num (int): The symbol identifier
        Returns:
            str: The retrieved symbol
        """
        return self.symbols[num]

    def items(self):
        """Returns all stored symbols
        Args:
            None
        Returns:
            dict:The included symbols
        """
        return self.symbols


class PythonDFA():
    """A DFA implementation that uses the
    same interface with python symautomata"""
    def __init__(self, alphabet=createalphabet()):
        """
        Args:
            alphabet (list): The imput alphabet
        Returns:
            None
        """
        self.states = []
        self.alphabet = alphabet
        self.nfa = False
        num = 1
        self.isyms = syms()
        self.osyms = syms()
        for char in alphabet + [EPSILON]:
            self.isyms.__setitem__(char, num)
            self.osyms.__setitem__(char, num)
            num = num + 1

    def __str__(self):
        """Describes DFA object"""
        return "This is a python DFA object with " + `len(self.states)` + " states"

    def __getitem__(self, state):
        """
        Retrieves a state
        Args:
            state (int): State identifier
        Returns:
            DFA state: The selected dfa state
        """
        return self.states[state]

    def __setitem__(self, state, newstate=None):
        """
        Sets a new state
        Args:
            state (int): State identifier
            newstate (DFA State): The selected dfa state
        Returns:
            None
        """
        self.states[state] = newstate

    def define(self):
        """If DFA is empty, create a sink state"""
        if len(self.states) == 0:
            for char in self.alphabet:
                self.add_arc(0, 0, char)
                self[0].final = False

    def add_state(self):
        """Adds a new state"""
        sid = len(self.states)
        self.states.append(DFAState(sid))
        return sid

    def add_arc(self, src, dst, char):
        """Adds a new Arc
        Args:
            src (int): The source state identifier
            dst (int): The destination state identifier
            char (str): The character for the transition
        Returns:
            None
        """
        # assert type(src) == type(int()) and type(dst) == type(int()), \
        #     "State type should be integer."
        # assert char in self.I
        #
        #print self.states
        #print src
        for s_idx in [src, dst]:
            if s_idx >= len(self.states):
                for i in range(len(self.states), s_idx + 1):
                    self.states.append(DFAState(i))
        for arc in self.states[src].arcs:
            if arc.ilabel == self.isyms.__getitem__(char) or char == EPSILON:
                self.nfa = True
                break
        self.states[src].arcs.append(
            DFAArc(src, dst, self.isyms.__getitem__(char)))

    def fixminimized(self, alphabet):
        """
        After pyfst minimization,
        all unused arcs are removed,
        and all sink states are removed.
        However this may break compatibility.
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        return
        # endstate = self.add_state()
        # for state in self.states:
        #     for char in alphabet:
        #         found = 0
        #         for arc in state.arcs:
        #             if self.isyms.find(arc.ilabel) == char:
        #                 found = 1
        #                 break
        #         if found == 0:
        #             self.add_arc(state.stateid, endstate, char)
        #
        # self[endstate].final = False
        #
        # for char in alphabet:
        #     self.add_arc(endstate, endstate, char)

    def _addsink(self, alphabet):
        """
        Adds a sink state
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        endstate = len(list(self.states))
        for state in self.states:
            for char in alphabet:
                found = 0
                for arc in state.arcs:
                    if self.isyms.find(arc.ilabel) == char:
                        found = 1
                        break
                if found == 0:
                    self.add_arc(state.stateid, endstate, char)

        self[endstate].final = False

        for char in alphabet:
            self.add_arc(endstate, endstate, char)

    def consume_input(self, inp):
        """
        Return True/False if the machine accepts/reject the input.
        Args:
            inp (str): input string to be consumed
        Returns:
            bool: A true or false value depending on if the DFA
                accepts the provided input
        """
        cur_state = sorted(
            self.states,
            key=attrgetter('initial'),
            reverse=True)[0]
        while len(inp) > 0:
            found = False
            for arc in cur_state.arcs:
                if self.isyms.find(arc.ilabel) == inp[0]:
                    cur_state = self[arc.nextstate]
                    inp = inp[1:]
                    found = True
                    break
            if not found:
                return False
        return cur_state.final

    def empty(self):
        """
        Return True if the DFA accepts the empty language.
        """
        # self.minimize()
        return len(list(self.states)) == 0

    def complement(self, alphabet):
        """
        Returns the complement of DFA
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        states = sorted(self.states, key=attrgetter('initial'), reverse=True)
        for state in states:
            if state.final:
                state.final = False
            else:
                state.final = True

    def init_from_acceptor(self, acceptor):
        """
        Adds a sink state
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        self.states = copy.deepcopy(acceptor.states)
        self.alphabet = copy.deepcopy(acceptor.alphabet)
        self.osyms = copy.deepcopy(acceptor.osyms)
        self.isyms = copy.deepcopy(acceptor.isyms)

    def save(self, txt_fst_file_name):
        """
        Save the machine in the openFST format in the file denoted by
        txt_fst_file_name.
        Args:
            txt_fst_file_name (str): The output file
        Returns:
            None
        """
        output_filename = open(txt_fst_file_name, 'w+')
        states = sorted(self.states, key=attrgetter('initial'), reverse=True)
        for state in states:
            for arc in state.arcs:
                itext = self.isyms.find(arc.ilabel)
                otext = self.osyms.find(arc.ilabel)
                output_filename.write(
                    '{}\t{}\t{}\t{}\n'.format(
                        state.stateid,
                        arc.nextstate,
                        itext.encode('hex'),
                        otext.encode('hex')))
            if state.final:
                output_filename.write('{}\n'.format(state.stateid))
        output_filename.close()

    def load(self, txt_fst_file_name):
        """
        Save the transducer in the text file format of OpenFST.
        The format is specified as follows:
            arc format: src dest ilabel olabel [weight]
            final state format: state [weight]
        lines may occur in any order except initial state must be first line
        Args:
            txt_fst_file_name (str): The input file
        Returns:
            None
        """
        with open(txt_fst_file_name, 'r') as input_filename:
            for line in input_filename:
                line = line.strip()
                split_line = line.split()
                if len(split_line) == 1:
                    self[int(split_line[0])].final = True
                else:
                    self.add_arc(int(split_line[0]), int(split_line[1]),
                                 split_line[2].decode('hex'))


    def minimize(self):
        """Minimizes the DFA using Hopcroft algorithm"""
        self.hopcroft()

    def intersect(self, other):
        """Constructs an unminimized DFA recognizing
        the intersection of the languages of two given DFAs.
        Args:
            other (DFA): The other DFA that will be used
                         for the intersect operation
        Returns:
        Returns:
            DFA: The resulting DFA
        """
        operation = bool.__and__
        self.cross_product(other, operation)
        return  self

    def __and__(self, other):
        """Constructs an unminimized DFA recognizing
               the intersection of the languages of two given DFAs.
               Args:
                   other (DFA): The other DFA that will be used
                                for the intersect operation
               Returns:
                   DFA: The resulting DFA
               """
        self.intersect(other)
        return self

    def symmetric_difference(self, other):
        """Constructs an unminimized DFA recognizing
        the symmetric difference of the languages of two given DFAs.
        Args:
            other (DFA): The other DFA that will be used
                         for the symmetric difference operation
        Returns:
            DFA: The resulting DFA
        """
        operation = bool.__xor__
        self.cross_product(other, operation)
        return  self

    def union(self, other):
        """Constructs an unminimized DFA recognizing the union of the languages of two given DFAs.
        Args:
            other (DFA): The other DFA that will be used
                         for the union operation
        Returns:
            DFA: The resulting DFA
        """
        operation = bool.__or__
        self.cross_product(other, operation)
        return self

    def __or__(self, other):
        """Constructs an unminimized DFA recognizing the union of the languages of two given DFAs.
                Args:
                    other (DFA): The other DFA that will be used
                                 for the union operation
                Returns:
                    DFA: The resulting DFA
                """
        self.union(other)
        return self


    def _epsilon_closure(self, state):
        """
        Returns the \epsilon-closure for the state given as input.
        """
        closure = set([state.stateid])
        stack = [state]
        while True:
            if not stack:
                break
            s = stack.pop()
            for arc in s:
                if self.isyms.find(arc.ilabel) != EPSILON or \
                        arc.nextstate in closure:
                    continue
                closure.add(arc.nextstate)
                stack.append(self.states[arc.nextstate])
        return closure


    def determinize(self):
        """
        Transforms a Non Deterministic DFA into a Deterministic
        Args:
            None
        Returns:
            DFA: The resulting DFA

        Creating an equivalent DFA is done using the standard algorithm.
        A nice description can be found in the book:
        Harry R. Lewis and Christos H. Papadimitriou. 1998.
        E
        print target_dfa_statelements of the Theory of Computation.
        """

        # Compute the \epsilon-closure for all states and save it in a diagram
        epsilon_closure = {}
        for state in self.states:
            sid = state.stateid
            epsilon_closure[sid] = self._epsilon_closure(state)

        # Get a transition diagram to speed up computations
        trans_table = {}
        for state in self.states:
            trans_table[state.stateid] = defaultdict(set)
            for arc in state:
                char = self.isyms.find(arc.ilabel)
                trans_table[state.stateid][char].add(arc.nextstate)

        # is_final function:
        # Given a set of nfa states representing a dfa_state return 1 if the
        # corresponding DFA state is a final state, i.e. if any of the
        # corresponding NFA states are final.
        is_final = lambda nfa_states, dfa_state: True \
            if sum([ int(nfa_states[x].final) for x in dfa_state ]) >= 1 \
            else False

        # Precomputation is over, start executing the conversion algorithm
        state_idx = 1
        nfa_states = copy.deepcopy(self.states)
        self.states = []
        # Initialize the new DFA state list
        self.add_state()
        new_initial = epsilon_closure[nfa_states[0].stateid]
        self.states[0].final = is_final(nfa_states, new_initial)

        dfa_state_idx_map = { frozenset(new_initial) : 0 }
        stack = [new_initial]
        while True:
            # Iterate until all added DFA states are processed.
            if not stack:
                break
            # This is a set of states from the NFA
            src_dfa_state = stack.pop()
            src_dfa_state_idx = dfa_state_idx_map[frozenset(src_dfa_state)]
            for char in self.alphabet:
                # Compute the set of target states
                target_dfa_state = set([])
                for nfa_state in src_dfa_state:
                    next_states = \
                        set([y for x in trans_table[nfa_state][char] \
                             for y in epsilon_closure[x] ])
                    target_dfa_state.update(next_states)
                # If the computed state set is not part of our new DFA add it,
                # along with the transition for the current character.
                if frozenset(target_dfa_state) not in dfa_state_idx_map:
                    self.add_state()
                    dfa_state_idx_map[frozenset(target_dfa_state)] = state_idx
                    self.states[state_idx].final = is_final(nfa_states,
                                                            target_dfa_state)
                    state_idx += 1
                    stack.append(target_dfa_state)

                dst_state_idx = dfa_state_idx_map[frozenset(target_dfa_state)]
                self.add_arc(src_dfa_state_idx, dst_state_idx, char)
        return self


    def invert(self):
        """Inverts the DFA final states"""
        for state in self.states:
            if state.final:
                state.final = False
            else:
                state.final = True

    def difference(self, other):
        """Performs the Diff operation between two atomata
        Args:
            other (DFA): The other DFA that will be used
                         for the difference operation
        Returns:
            DFA: The resulting DFA
        """
        other.invert()
        self.intersect(other)
        return self

    def __sub__(self, other):
        """Performs the Diff operation between two atomata
        Args:
            other (DFA): The other DFA that will be used
                        for the difference operation
        Returns:
            DFA: The resulting DFA
        """
        self.difference(other)
        return self


    def hopcroft(self):
        """
        Performs the Hopcroft minimization algorithm
        Args:
            None
        Returns:
            DFA: The minimized input DFA
        """

        def _getset(testset, partition):
            """
            Checks if a set is in a partition
            Args:
                testset (set): The examined set
                partition (list): A list of sets
            Returns:
                bool: A value indicating if it is a member or not
            """
            for part in partition:
                if set(testset) == set(part):
                    return True
            return None

        def _create_transitions_representation(graph):
            """
            In order to speedup the transition iteration using
            the alphabet, the function creates an index
            Args:
                graph (DFA): The input dfa
                state (DFA state): The examined state
            Returns:
                dict: The generated transition map
            """
            return {x.stateid:{self.isyms.find(arc.ilabel): arc.nextstate \
                               for arc in x} for x in graph.states}

        def _create_reverse_transitions_representation(graph):
            """
            In order to speedup the transition iteration using
            the alphabet, the function creates an index
            Args:
                graph (DFA): The input dfa
                state (DFA state): The examined state
            Returns:
                dict: The generated transition map
            """
            return {x.stateid: {self.isyms.find(arc.ilabel): arc.nextstate \
                                for arc in x} for x in graph.states}

        def _reverse_to_source(target, group1):
            """
            Args:
                target (dict): A table containing the reverse transitions for each state
                group1 (list): A group of states
            Return:
                Set: A set of states for which there is a transition with the states of the group
            """
            new_group = []
            for dst in group1:
                new_group += target[dst]
            return set(new_group)

        def _get_group_from_state(groups, sid):
            """
            Args:
                sid (int): The state identifier
            Return:
                int: The group identifier that the state belongs
            """
            for index, selectgroup in enumerate(groups):
                if sid in selectgroup:
                    return index

        def _delta(graph, cur_state, char):
            """
            Function describing the transitions
            Args:
                graph (DFA): The DFA states
                cur_state (DFA state): The DFA current state
                char (str):: The char that will be used for the transition
            Return:
                DFA Node: The next state
            """
            for arc in cur_state.arcs:
                if graph.isyms.find(arc.ilabel) == char:
                    return graph[arc.nextstate]

        def _partition_group(bookeeping, group):
            """
            Args:
                group (list):  A group of states
            Return:
                tuple: A set of two groups
            """
            for (group1, group2) in bookeeping:
                if group & group1 != set() and not group.issubset(group1):
                    new_g1 = group & group1
                    new_g2 = group - group1
                    return (new_g1, new_g2)
                if group & group2 != set() and not group.issubset(group2):
                    new_g1 = group & group2
                    new_g2 = group - group2
                    return (new_g1, new_g2)
            assert False, "Unmatched group partition"

        def _object_set_to_state_list(objectset):
            """
            Args:
                objectset (list): A list of all the DFA states (as objects)
            Return:
                list: A list of all the DFA states (as identifiers)
            """
            return [state.stateid for state in objectset]

        def _get_accepted(graph):
            """
            Find the accepted states
            Args:
                graph (DFA): The DFA states
            Return:
                list: Returns the list of the accepted states
            """
            return [state for state in graph \
                    if  state.final != TropicalWeight(float('inf'))]

        graph = self

        # Find Q
        set_q = set(_object_set_to_state_list(graph.states))
        # We will work with states addresses here instead of states stateid for
        # more convenience
        set_f = set(_object_set_to_state_list(_get_accepted(graph)))
        # Perform P := {F, Q-F}
        set_nf = set_q.copy() - set_f.copy()
        groups = [set_f.copy(), set_nf.copy()]
        bookeeping = [(set_f, set_nf)]

        done = False
        while not done:
            done = True
            new_groups = []
            for selectgroup in groups:
                # _check for each letter if it splits the current group
                for character in self.alphabet:
                    # print 'Testing symbol: ', c
                    target = defaultdict(list)
                    target_states = defaultdict(int)
                    new_g = [set(selectgroup)]
                    for sid in selectgroup:
                        # _check if all transitions using c are going in a state
                        # in the same group. If they are going on a different
                        # group then split
                        deststate = _delta(graph, graph[sid], character)
                        destgroup = _get_group_from_state(groups,
                            deststate.stateid)
                        target[destgroup].append(sid)
                        target_states[destgroup] = deststate.stateid
                    if len(target) > 1:

                        inv_target_states = {
                            v: k for k, v in target_states.iteritems()}
                        new_g = [set(selectedstate) for selectedstate in target.values()]
                        done = False
                        # Get all the partitions of destgroups
                        queue = [set([x for x in target_states.values()])]
                        while queue:
                            top = queue.pop(0)
                            (group1, group2) = _partition_group(bookeeping, top)
                            ng1 = _reverse_to_source(
                                target, [inv_target_states[x] for x in group1])
                            ng2 = _reverse_to_source(
                                target, [inv_target_states[x] for x in group2])

                            bookeeping.append((ng1, ng2))

                            if len(group1) > 1:
                                queue.append(group1)
                            if len(group2) > 1:
                                queue.append(group2)
                        break
                new_groups += new_g

            # End of iteration for the k-equivalence
            # Assign new groups and check if any change occured
            groups = new_groups

        # Make a copy of the old states, and prepare the
        # automaton to host the minimum states

        oldstates = copy.deepcopy(self.states)
        self.states = []
        self.define()

        def findpart(stateid, partitions):
            """Searches for the groupt that the state identifier
            belongs to.
            Args:
                stateid (int): The state identifier
                partitions (list): The list of the groups
            Returns:
                set: The group that the stateid belongs to.
            """
            for group in partitions:
                if stateid in group:
                    return frozenset(group)
            return frozenset(set(            ))

        def add_state_if_not_exists(group, statesmap, final):
            """
            Adds a new state in the final dfa. It initialy checks if
            the group of states is already registered to the automaton.
            If it is registered, the state identifier is returned, or
            else, a new state is added.
            Args:
                group (frozenset):  The group that the state identifier belongs
                statesmap (dict):   A dictionary that maintains the state
                                    identifiers for each forzenset
                final (bool):       A value indicating if the current state is
                                    final
            Returns:
                int: The new state identifier
            """
            if group not in statesmap:
                sid = self.add_state()
                self[sid].final = final
                statesmap[group] = sid
            return statesmap[group]

        statesmap = {}
        self.states = []
        group = findpart(0, groups)
        sid = add_state_if_not_exists(frozenset(list(group)), statesmap,
                                      oldstates[0].final)
        self[sid].initial = True
        for group in groups:
            if len(group) == 0:
                continue
            sid = add_state_if_not_exists(frozenset(group), statesmap,
                                          oldstates[list(group)[0]].final)
            state = next(iter(group))
            for arc in oldstates[state]:
                dst_group = findpart(arc.nextstate, groups)
                dst_sid = add_state_if_not_exists(
                    dst_group, statesmap, oldstates[arc.nextstate].final)
                self.add_arc(sid, dst_sid, graph.isyms.find(arc.ilabel))


    def cross_product(self, dfa_2, accept_method):
        """A generalized cross-product constructor over two DFAs.
        The third argument is a binary boolean function f; a state (q1, q2) in the final
        DFA accepts if f(A[q1],A[q2]), where A indicates the acceptance-value of the state.
        Args:
            dfa_2: The second dfa
            accept_method: The boolean action
        Returns:
            None
        """
        dfa_1states = copy.deepcopy(self.states)
        dfa_2states = dfa_2.states
        self.states = []
        states = {}


        def _create_transitions_representation(graph, state):
            """
            In order to speedup the transition iteration using
            the alphabet, the function creates an index
            Args:
                graph (DFA): The input dfa
                state (DFA state): The examined state
            Returns:
                dict: The generated transition map
            """
            return {self.isyms.find(arc.ilabel): graph[arc.nextstate] for arc in state}


        def _add_state_if_nonexistent(state_a, state_b):
            """
            Adds a new state in the final dfa, which is the
            combination of the input states. The initial and final
            flag is also placed on the new state. If the state already
            exists, its identifier is being returned.
            Args:
                state_a: The fist state identifier
                state_b: The second state identifier
            Returns:
                int: The new state identifier
            """
            if (state_a.stateid, state_b.stateid) not in states:
                states[(state_a.stateid, state_b.stateid)] \
                    = self.add_state()
                self[states[(state_a.stateid, state_b.stateid)]].initial \
                    = state_a.initial and state_b.initial
                self[states[(state_a.stateid, state_b.stateid)]].final \
                    = accept_method(state_a.final, state_b.final)
            return states[(state_a.stateid, state_b.stateid)]

        for state1, state2 in product(dfa_1states, dfa_2states):
            sid1 = _add_state_if_nonexistent(state1, state2)
            transitions_s1 = _create_transitions_representation(dfa_1states, state1)
            transitions_s2 = _create_transitions_representation(dfa_2states, state2)
            for char in self.alphabet:
                sid2 = _add_state_if_nonexistent(
                    transitions_s1[char], transitions_s2[char])
                self.add_arc(sid1, sid2, char)




"""This module creates a PDA from a CNF """
from pda import PDA, PDAState


class ProdStruct:
    """This class is used as a struct for each CNF rule"""
    type = 0
    a = 0
    b0 = 0
    b1 = 0

    def __init__(self):
        """ Initialization class """
        pass


class CnfPda:
    """This class creates a PDA from a CNF"""
    def __init__(self, alphabet):
        """
        Args:
            alphabet (list): The input alphabet
        """
        self.alphabet = alphabet


    def _mkpda(self, nonterms, productions, productions_struct, terminals, splitstring=1):
        """
        This function generates a PDA from a CNF grammar as described in:
          -  http://www.oit.edu/faculty/sherry.yang/CST229/Lectures/7_pda.pdf
          -  http://www.eng.utah.edu/~cs3100/lectures/l18/pda-notes.pdf

        If all of the grammar productions are in the Chomsky Normal Form,
        then follow the template for constructing a pushdown symautomata:

        1. Start
        2. Push S
        3. Pop
        4. Case:
            Nonterminal A: For every production rule of this form: A: BC, Push C and then Push B

        Args:
             nonterms (list): Non terminals list
             productions (dict): productions in the CNF form:
                                    A -> a or A -> b0b1, or S -> e
             productions_struct (dict):  productions in the CNF form in structure form
                                    object.a for A -> a,
                                    object.b0 and object.b1 for A -> b0b1
                                    and object.type where type is
                                        1 for A-->a and 2 for A-->b0b1
             terminals (list): All terminals
             splitstring (bool): If enabled an extra space is added after each symbol.
        Returns:
            PDA: The generated PDA
        """
        pda = PDA(self.alphabet)
        pda.nonterminals = nonterms
        pda.terminals = terminals

        pda.s[pda.n] = PDAState()
        pda.s[pda.n].id = pda.n
        pda.s[pda.n].sym = '@closing'
        pda.s[pda.n].type = 1
        pda.s[pda.n].trans[1] = [0]

        pda.n = pda.n + 1
        pda.s[pda.n] = PDAState()
        pda.s[pda.n].id = pda.n
        pda.s[pda.n].type = 1
        pda.s[pda.n].sym = nonterms[0]
        pda.s[pda.n].trans[2] = [0]

        pda.n = pda.n + 1
        pda.s[pda.n] = PDAState()
        pda.s[pda.n].id = pda.n
        pda.s[pda.n].type = 2
        pda.s[pda.n].trans[0] = ['@closing']

        counter = 0
        i = 0
        while i < len(nonterms):
            j = 0
            while j < len(productions[nonterms[i]]):
                if productions_struct[counter].type == 1:
                    # ADD AND CONNECT STATE
                    pda.n = pda.n + 1
                    pda.s[pda.n] = PDAState()
                    pda.s[pda.n].id = pda.n
                    if pda.n not in pda.s[2].trans:
                        pda.s[2].trans[pda.n] = []
                    pda.s[2].trans[pda.n].append(nonterms[i])

                    if splitstring == 0:

                        # FILL NEW STATE  READ
                        pda.s[pda.n].type = 3
                        pda.s[pda.n].trans[2] = [productions_struct[counter].a]

                    else:
                        # THE FOLLOWIN SWITCH IS DUE TO THE REQUIREMENT OF
                        # HAVING STRINGS SPLITTED TO SYMBOLS AND CAN INTERSECT
                        # WITH DFA

                        if productions_struct[counter].a not in terminals or \
                                        len(productions_struct[counter].a) == 1:
                            # FILL NEW STATE  READ
                            pda.s[pda.n].type = 3
                            pda.s[pda.n].trans[pda.n + 1] = [productions_struct[counter].a.lower()]
                            pda.n = pda.n + 1
                            pda.s[pda.n] = PDAState()
                            pda.s[pda.n].id = pda.n
                            pda.s[pda.n].type = 3
                            pda.s[pda.n].trans[2] = [' ']

                        else:
                            pda.s[pda.n].type = 3
                            pda.s[pda.n].trans[pda.n + 1] = \
                                [productions_struct[counter].a[0].lower()]
                            k = 1
                            while k < len(productions_struct[counter].a) - 1:
                                pda.n = pda.n + 1
                                pda.s[pda.n] = PDAState()
                                pda.s[pda.n].id = pda.n
                                pda.s[pda.n].type = 3
                                pda.s[pda.n].trans[pda.n +1] = \
                                    [productions_struct[counter].a[k].lower()]
                                k = k + 1
                            pda.n = pda.n + 1
                            pda.s[pda.n] = PDAState()
                            pda.s[pda.n].id = pda.n
                            pda.s[pda.n].type = 3

                            pda.s[pda.n].trans[pda.n + 1] = \
                                [productions_struct[counter].a[-1].lower()]
                            pda.n = pda.n + 1
                            pda.s[pda.n] = PDAState()
                            pda.s[pda.n].id = pda.n
                            pda.s[pda.n].type = 3
                            pda.s[pda.n].trans[2] = [' ']

                else:
                    # ADD AND CONNECT PUSH STATE
                    pda.n = pda.n + 1
                    pda.s[pda.n] = PDAState()
                    pda.s[pda.n].id = pda.n
                    if pda.n not in pda.s[2].trans:
                        pda.s[2].trans[pda.n] = []
                    pda.s[2].trans[pda.n].append(nonterms[i])

                    # FILL NEW STATE
                    pda.s[pda.n].type = 1
                    pda.s[pda.n].sym = productions_struct[counter].b1
                    pda.s[pda.n].trans[(pda.n) + 1] = [0]

                    # ADD AND CONNECT PUSH STATE (ALREADY CONNECTED)
                    pda.n = pda.n + 1
                    pda.s[pda.n] = PDAState()
                    pda.s[pda.n].id = pda.n

                    # FILL NEW STATE
                    pda.s[pda.n].type = 1
                    pda.s[pda.n].sym = productions_struct[counter].b0
                    pda.s[pda.n].trans[2] = [0]

                j = j + 1
                counter = counter + 1

            i = i + 1
        return pda

    def initialize(
            self,
            nonterminal,
            productions,
            terminals,
            splitstring=1):
        """

        productions in structure form <1> A-->b , <2> a-->BC
        """
        cnt = 0
        i = 0
        productions_struct = {}
        while i < len(nonterminal):
            j = 0
            while j < len(productions[nonterminal[i]]):
                productions_struct[cnt] = ProdStruct()
                if 'a' in productions[nonterminal[i]][j]:
                    productions_struct[cnt].type = 1
                    productions_struct[cnt].a = productions[nonterminal[i]][j]['a']

                else:
                    productions_struct[cnt].type = 2
                    productions_struct[cnt].b0 = productions[nonterminal[i]][j]['b0']
                    productions_struct[cnt].b1 = productions[nonterminal[i]][j]['b1']
                cnt = cnt + 1
                j = j + 1
            i = i + 1
        return self._mkpda(
            nonterminal,
            productions,
            productions_struct,
            terminals,
            splitstring)



"""This module generates a CNF grammar from a PDA object"""
from sys import argv

from alphabet import createalphabet
from cfggenerator import CFGGenerator
from cfgpda import CfgPDA
from pda import PDAState
from FAdo import cfg

class IntersectionHandling():
    """
    After the intersection (product operation) there would be final
    transitions to non accepted states (according to the DFA
    accepted states). These transitions should be removed before
    generating the CFG
    """

    def __init__(self):
        pass

    def get(self, statediag, dfaaccepted):
        """
        # - Remove all the POP (type - 2) transitions to state 0,non DFA accepted
        # for symbol @closing
        # - Generate the accepted transitions
        - Replace DFA accepted States with a push - pop symbol and two extra states
        Args:
            statediag (list): The states of the PDA
            dfaaccepted (list):The list of DFA accepted states
        Returns:
            list: A cleaned, smaller list of DFA states
        """

        newstatediag = {}

        newstate = PDAState()
        newstate.id = 'AI,I'  # BECAREFUL WHEN SIMPLIFYING...
        newstate.type = 1
        newstate.sym = '@wrapping'
        transitions = {}
        transitions[(0, 0)] = [0]
        newstate.trans = transitions
        i = 0
        newstatediag[i] = newstate
        # print 'accepted:'
        # print dfaaccepted
        for stateid in statediag:
            state = statediag[stateid]
            # print state.id
            if state.type == 2:
                for state2id in dfaaccepted:
                    # print state.id[1]
                    if state.id[1] == state2id:
                        # print 'adding...'
                        state.trans['AI,I'] = ['@wrapping']
                        # print state.trans
                        break
            i = i + 1
            newstatediag[i] = state
        return newstatediag


class ReducePDA():
    """Use BFS to search for unreachable states and remove them"""
    def __init__(self):
        pass

    def bfs(self, graph, start):
        """
        Performs BFS operation for eliminating useless loop transitions
        Args:
            graph (PDA): the PDA object
            start (PDA state): The PDA initial state
        Returns:
            list: A cleaned, smaller list of DFA states
        """
        newstatediag = {}

        # maintain a queue of paths
        queue = []
        visited = []
        # push the first path into the queue
        queue.append(start)
        while queue:
            # get the first path from the queue
            state = queue.pop(0)
            # get the last node from the path
            # visited
            visited.append(state.id)
            # enumerate all adjacent nodes, construct a new path and push it
            # into the queue
            for key in state.trans:
                if state.trans[key] != []:
                    if key not in visited:
                        for nextstate in graph:
                            if graph[nextstate].id == key:
                                queue.append(graph[nextstate])
                                break
        i = 0
        for state in graph:
            if graph[state].id in visited:
                newstatediag[i] = graph[state]
                i = i + 1

        return newstatediag

    def get(self, statediag):
        """
        Args:
            statediag (list): The states of the PDA
        Returns:
            list: A reduced list of states using BFS
        """
        if len(statediag) < 1:
            print 'PDA is empty and can not be reduced'
            return statediag
        newstatediag = self.bfs(statediag, statediag[0])
        return newstatediag


class SimplifyStateIDs():
    """
    Transform state IDs to a more simple form (sequencial numbers).
    Should be used after product operation (intersection)
    """

    def __init__(self):
        pass

    def get(self, statediag, accepted=None):
        """
        Replaces complex state IDs as generated from the product operation,
        into simple sequencial numbers. A dictionaty is maintained in order
        to map the existed IDs.
        Args:
            statediag (list): The states of the PDA
            accepted (list): the list of DFA accepted states
        Returns:
            list:
        """
        count = 0
        statesmap = {}
        newstatediag = {}
        for state in statediag:

            # Simplify state IDs
            if statediag[state].id not in statesmap:
                statesmap[statediag[state].id] = count
                mapped = count
                count = count + 1
            else:
                mapped = statesmap[statediag[state].id]

            # Simplify transitions IDs

            transitions = {}
            for nextstate in statediag[state].trans:
                if nextstate not in statesmap:
                    statesmap[nextstate] = count
                    transmapped = count
                    count = count + 1
                else:
                    transmapped = statesmap[nextstate]
                transitions[transmapped] = statediag[state].trans[nextstate]
            newstate = PDAState()
            newstate.id = mapped
            newstate.type = statediag[state].type
            newstate.sym = statediag[state].sym
            newstate.trans = transitions
            newstatediag[mapped] = newstate
        newaccepted = None
        if accepted is not None:
            newaccepted = []
            for accepted_state in accepted :
                if (0, accepted_state) in statesmap:
                    newaccepted.append(statesmap[(0, accepted_state)])
        return newstatediag, count, newaccepted


class ReadReplace():
    """
    Removes all READ (type - 3) states and replaces them with PUSH (type - 1) and POP (type - 2).
    Should be used before PDA to CFG operation.
    """

    statediag = []
    quickresponse = {}
    quickresponse_types = {}
    toadd = []
    biggestid = 0

    def __init__(self, statediag=[], thebiggestid=None):
        """
        Find the biggest State ID
        Args:
            statediag (list): The states of the PDA
            thebiggestid (int): The binggest state identifier
        Returns:
            None
        """
        self.statediag = []
        self.quickresponse = {}
        self.quickresponse_types = {}
        self.toadd = []
        self.biggestid = 0

        if thebiggestid is None:
            for state in statediag:
                if statediag[state].id > self.biggestid:
                    self.biggestid = statediag[state].id
        else:
            self.biggestid = thebiggestid
        self.statediag = statediag

    def nextstate(self):
        """
        Always return the biggest state ID + 1
        """
        self.biggestid = self.biggestid + 1
        return self.biggestid

    def _generate_state(self, trans):
        """
        Creates a new POP state (type - 2) with the same transitions.
        The POPed symbol is the unique number of the state.
        Args:
            trans (dict): Transition dictionary
        Returns:
            Int: The state identifier
        """
        state = PDAState()
        state.id = self.nextstate()
        state.type = 2
        state.sym = state.id
        state.trans = trans.copy()
        self.toadd.append(state)
        return state.id

    def replace_read(self):
        """
        Replaces all READ (type - 3) states to a PUSH (type - 1) and a POP (type - 2).
        The actual state is replaced with the PUSH, and a new POP is created.
        """
        for statenum in self.statediag:
            state = self.statediag[statenum]
            if state.type == 3:  # READ state
                state.type = 1
                destination_and_symbol = self._generate_state(state.trans)
                state.sym = destination_and_symbol
                state.trans = {}
                state.trans[destination_and_symbol] = [0]
        statenumber_identifier = len(self.statediag) + 1
        for state in self.toadd:
            self.statediag[statenumber_identifier] = state
            statenumber_identifier = statenumber_identifier + 1
        return self.statediag


class PdaCnf():
    """This class manages PDA to CNF generation"""
    rules = []
    statediag = []
    accepted = []

    def insert_start_to_accepting(self):
        """
        Insert the start rule S -> A0,0
            or alphabet -> AI,I (IF intersected, the DFA Accepted states
            will be eliminated by adding a wrapping state with push pop symbol)
        """
        self.rules.append('S: A0,0')

    def insert_self_to_empty_and_insert_all_intemediate(self, optimized):
        """
        For each state qi of the PDA, we add the rule Aii -> e
        For each triplet of states qi, qj and qk, we add the rule Aij -> Aik Akj.
        Args:
            optimized (bool): Enable or Disable optimization - Do not produce O(n^3)
        """
        for state_a in self.statediag:
            self.rules.append('A' +repr(state_a.id) +',' + repr(state_a.id) + ': @empty_set')

            # If CFG is not requested, avoid the following O(n^3) rule.
            # It can be solved and a string can be generated faster with BFS of DFS

            if optimized == 0:
                for state_b in self.statediag:
                    if state_b.id != state_a.id:
                        for state_c in self.statediag:
                            if state_c.id != state_a.id \
                                    and state_b.id != state_c.id:
                                self.rules.append('A' + repr(state_a.id)
                                                  + ',' + repr(state_c.id)
                                                  + ': A' + repr(state_a.id)
                                                  + ',' + repr(state_b.id)
                                                  + ' A' + repr(state_b.id)
                                                  + ',' + repr(state_c.id)
                                                  + '')

    def insert_symbol_pushpop(self):
        """
        For each stack symbol t E G, we look for a pair of states, qi and qj,
        such that the PDA in state qi can read some input a E S and push t
        on the stack and in state state qj can read some input b E S and pop t
        off the stack. In that case, we add the rule Aik -> a Alj b
        where (ql,t) E d(qi,a,e) and (qk,e) E d(qj,b,t).
        """

        for state_a in self.statediag:
            if state_a.type == 1:
                found = 0
                for state_b in self.statediag:
                    if state_b.type == 2 and state_b.sym == state_a.sym:
                        found = 1
                        for j in state_a.trans:
                            if state_a.trans[j] == [0]:
                                read_a = ''
                            else:
                                new = []
                                for selected_transition in state_a.trans[j]:
                                    if selected_transition == ' ':
                                        new.append('&')
                                    else:
                                        new.append(selected_transition)

                                read_a = " | ".join(new)
                            for i in state_b.trans:
                                if state_b.trans[i] == [0]:
                                    read_b = ''
                                else:
                                    new = []
                                    for selected_transition in state_b.trans[i]:
                                        if selected_transition == ' ':
                                            new.append('&')
                                        else:
                                            new.append(selected_transition)
                                    read_b = " | ".join(new)
                                self.rules.append(
                                    'A' + repr(state_a.id)
                                    + ',' + repr(i)
                                    + ':' + read_a
                                    + ' A' + repr(j)
                                    + ',' + repr(state_b.id)
                                    + ' ' + read_b)
                if found == 0:

                    # A special case is required for State 2, where the POPed symbols
                    # are part of the transitions array and not defined for "sym" variable.

                    for state_b in self.statediag:
                        if state_b.type == 2 and state_b.sym == 0:
                            for i in state_b.trans:
                                if state_a.sym in state_b.trans[i]:
                                    for j in state_a.trans:
                                        if state_a.trans[j] == [0]:
                                            read_a = ''
                                        else:
                                            read_a = " | ".join(
                                                state_a.trans[j])
                                        self.rules.append(
                                            'A' + repr(state_a.id)
                                            + ',' + repr(i)
                                            + ':' + read_a
                                            + ' A' + repr(j)
                                            + ',' + repr(state_b.id))
                                        # print
                                        # 'A'+`state_a.id`+','+`i`+':'+read_a+'
                                        # A'+`j`+','+`state_b.id`
                                        found = 1
                if found == 0:
                    print "ERROR: symbol " + repr(state_a.sym) \
                          + ". It was not found anywhere in the graph."

    def get_rules(self, optimized):
        """
        Args:
            optimized (bool): Enable or Disable optimization - Do not produce O(n^3)
        Return:
            list: The CFG rules
        """
        self.insert_start_to_accepting()
        # If CFG is not requested, avoid the following O(n^3) rule.
        # It can be solved and a string can be generated faster with BFS of DFS

        if optimized == 0:
            self.insert_self_to_empty_and_insert_all_intemediate(optimized)
        self.insert_symbol_pushpop()
        return self.rules

    def __init__(self, states, dfaaccepted=[]):
        self.rules = []
        self.statediag = []
        self.accepted = []
        self.accepted = dfaaccepted
        for key in states:
            self.statediag.append(states[key])


def _read_file(fname):
    """
    Args:
        fname (str): Name of the grammar file to be parsed
    Return:
        list: The grammar rules
    """
    with open(fname) as input_file:
        re_grammar = [x.strip('\n') for x in input_file.readlines()]
    return re_grammar


def main():
    """
    Function for PDA to CNF Operation
    :type argv: list
    :param argv: Parameters
    """
    if len(argv) < 3:
        print 'Usage for getting CFG: %s CFG_fileA CFG ' % argv[0]
        print 'Usage for getting STR: %s CFG_fileA STR ' \
              'Optimize[0 or 1] splitstring[0 or 1] ' % argv[0]
        print ''
        print 'For example: python pdacnf.py grammar.y STR 1 0'
        print '             python pdacnf.py grammar.y STR 1 1'
        print '             python pdacnf.py grammar.y CFG'
        return

    alphabet = createalphabet()

    mode = argv[2]

    optimized = 0
    splitstring = 0
    if mode == 'STR':
        optimized = int(argv[3])
        splitstring = int(argv[4])

    cfgtopda = CfgPDA(alphabet)
    print '* Parsing Grammar:',
    mma = cfgtopda.yyparse(argv[1])
    print 'OK'
    print ' - Total PDA states are ' + repr(len(mma.s))

    print '* Simplify State IDs:',
    simple_a = SimplifyStateIDs()
    mma.s, biggestid, newaccepted = simple_a.get(mma.s)
    if newaccepted:
        print 'OK'
    else:
        print 'OK'

    print '* Eliminate READ states:',
    replace = ReadReplace(mma.s, biggestid)
    mma.s = replace.replace_read()
    print 'OK'
    print ' - Total PDA states now are ' + repr(len(mma.s))
    maxstate = replace.nextstate() - 1

    print '* Reduce PDA:',
    simple_b = ReducePDA()
    mma.s = simple_b.get(mma.s)
    print 'OK'
    print ' - Total PDA states now are ' + repr(len(mma.s))

    print '* PDA to CFG transformation:',
    cnfgenerator = PdaCnf(mma.s)
    grammar = cnfgenerator.get_rules(optimized)
    print 'OK'
    print ' - Total CFG rules generated: ' + repr(len(grammar))

    if mode == 'STR':
        gen = CFGGenerator(cfg.CNF(cfg.gRules(grammar, ":")),
                           optimized=optimized,
                           splitstring=splitstring,
                           maxstate=maxstate)
        print gen.generate()
    else:
        print grammar


    # For example, for given cfg:
    #
    #     S: main
    #     main: 1 mainnew 1
    #     mainnew:  u
    #     mainnew:  main
    #
    # the generated grammar (optimized) would be:
    #
    # ['S: A0,0', 'A0,0: A1,2', 'A1,3: A2,2', 'A3,9: A11,2',
    # 'A4,8: A12,2', 'A5,2: A18,18 u', 'A6,7: A13,2',
    # 'A7,10: A14,2', 'A8,10: A15,2', 'A9,10: A16,2',
    # 'A10,2: A19,19 1', 'A11,10: A2,2', 'A12,10: A2,2',
    #  'A13,10: A2,2', 'A14,5: A2,2', 'A14,6: A2,2',
    # 'A15,5: A2,2', 'A15,6: A2,2', 'A16,5: A2,2', 'A16,6: A2,2']
    #
    # and can be solved as following:
    #
    #     S: A0,0
    #     S: A1,2
    #     S: (A1,3 A3,2)
    #     S: A2,2 (A3,9 A9,2)
    #     S: A11,2 (A9,10 A10,2)
    #     S: (A11,10 A10,2) (A16,2 A19,19 1)
    #     S: (A2,2  A19,19 1) ((A16,5 A5,2) 1)
    #     S:  1 (A2,2 (A18,18 u)) 1
    #     S: 1u1



if __name__ == '__main__':
    main()



"""
This module performs all basic PDA operations.
It is an interface for push down automata.
"""


class PDAState(object):
    """This is the structure for a PDA state"""
    type = 0
    sym = 0
    trans = None
    id = 0

    def printer(self):
        """Prints PDA state attributes"""
        print " ID " + repr(self.id)
        if self.type == 0:
            print " Tag: - "
            print " Start State - "
        elif self.type == 1:
            print " Push " + repr(self.sym)
        elif self.type == 2:
            print " Pop State " + repr(self.sym)
        elif self.type == 3:
            print " Read State " + repr(self.sym)
        elif self.type == 4:
            print " Stop State " + repr(self.sym)
        for j in self.trans:
            if len(self.trans[j]) > 1 or (len(self.trans[j]) == 1):
                for symbol in self.trans[j]:
                    print " On Symbol " + repr(symbol) + " Transition To State " + repr(j)

    def __init__(self):
        """State Initialization"""
        self.trans = {}


class syms:
    """The DFA accepted symbols"""

    def __init__(self):
        """Initialize symbols"""
        self.symbols = {}
        self.reversesymbols = {}

    def __getitem__(self, char):
        """
        Finds a symbol identifier based on the input character
        Args:
            char (str): The symbol character
        Returns:
            int: The retrieved symbol identifier
        """
        return self.reversesymbols[char]

    def __setitem__(self, char, num):
        """
        Sets a symbol
        Args:
            char (str): The symbol character
            num (int): The  symbol identifier
        Returns:
           None
        """
        self.symbols[num] = char
        self.reversesymbols[char] = num

    def find(self, num):
        """
        Finds a symbol based on its identifier
        Args:
            num (int): The symbol identifier
        Returns:
            str: The retrieved symbol
        """
        return self.symbols[num]

    def items(self):
        """Returns all stored symbols
        Args:
            None
        Returns:
            dict:The included symbols
        """
        return self.symbols

class PythonPDA(object):
    """This is the structure for a PDA"""
    n = 0
    s = None
    accepted = None
    terminals = None
    nonterminals = None


    def printer(self):
        """Prints PDA states and their attributes"""
        i = 0
        while i < self.n + 1:
            print "--------- State No --------" + repr(i)
            self.s[i].printer()
            i = i + 1

    def consume_input(self, mystr, stack=[], state=1, curchar=0, depth=0):
        """
        Consumes an input and validates if it is accepted
        Args:
            mystr (str): the input string to be consumes
            stack (list): the stack of symbols
            state (int): the current state of the PDA
            curchar (int): the index of the consumed character
            depth (int): the depth of the function call in the stack
        Returns:
            bool: A value indicating the correct or erroneous execution
        """
        mystrsplit = mystr.split(' ')
        if self.s[state].type == 1:
            stack.append(self.s[state].sym)
            if len(self.s[state].trans) > 0:
                state = self.s[state].trans[0]
                if self.parse(
                        mystr,
                        stack=stack,
                        state=state,
                        curchar=curchar,
                        depth=depth + 1) == 1:
                    return True
            return False
        if self.s[state].type == 2:
            if len(stack) == 0:
                return False
            sym = stack.pop()
            for key in self.s[state].trans:
                if sym in self.s[state].trans[key]:
                    if self.parse(
                            mystr,
                            stack=stack,
                            state=key,
                            curchar=curchar,
                            depth=depth + 1) == 1:
                        return True
            return False
        if self.s[state].type == 3:
            for key in self.s[state].trans:
                if mystrsplit[curchar] in self.s[state].trans[key]:
                    # print 'found '
                    if curchar + 1 == len(mystrsplit) \
                            and 'closing' in self.s[key].trans:
                        return True
                    elif curchar + 1 == len(mystrsplit):
                        return False

                    # print 'lets try as next state the state ' + repr(key)
                    if self.parse(
                            mystr,
                            stack=stack,
                            state=key,
                            curchar=curchar + 1,
                            depth=depth + 1) == 1:
                        return True
            return False


    def __init__(self, alphabet):
        """
        Args:
            alphabet (list): the input alphabet
        Returns:
            None
        """
        self.s = {}
        self.alphabet = alphabet

        num = 1
        self.isyms = syms()
        self.osyms = syms()
        for char in alphabet:
            self.isyms.__setitem__(char, num)
            self.osyms.__setitem__(char, num)
            num = num + 1


import os.path
"""This module configures that alphabet."""


def _load_alphabet(filename):
    """
    Load a file containing the characters of the alphabet.
    Every unique character contained in this file will be used as a symbol
    in the alphabet.
    """
    with open(filename, 'r') as f:
        return list(set(f.read()))

def createalphabet(alphabetinput=None):
    """
    Creates a sample alphabet containing printable ASCII characters
    """
    if alphabetinput and os.path.isfile(alphabetinput):
        return _load_alphabet(alphabetinput)
    elif alphabetinput:
        alpha = []
        setlist = alphabetinput.split(',')
        for alphaset in setlist:
            a = int(alphaset.split('-')[0])
            b = int(alphaset.split('-')[1])
            for i in range(a, b):
                alpha.append(str(unichr(i)))
        return alpha

    alpha = []
    for i in range(58, 64):
        alpha.append(str(unichr(i)))
    for i in range(32, 57):
        alpha.append(str(unichr(i)))
    for i in range(65, 126):
        alpha.append(str(unichr(i)))
    return alpha


"""
This module transforms a pyfst DFA to regular expressions
using the Brzozowski Algebraic Method
"""
import sys
import re
from collections import OrderedDict
from operator import attrgetter

from alphabet import createalphabet
from dfa import DFA
from flex2fst import Flexparser


class Regex:
    """
    This class transforms a pyfst DFA to regular expressions
    using the Brzozowski Algebraic Method
    """

    def __init__(self, input_fst_a, alphabet=None):
        """
        Args:
            input_fst_a (DFA): The DFA states
            alphabet (list): the input alphabet
        Returns:
            None
        """
        if alphabet is None:
            alphabet = createalphabet()
        self.alphabet = alphabet
        self.mma = DFA(self.alphabet)
        self.mma.init_from_acceptor(input_fst_a)
        # a is a matrix that holds all transitions
        # If there is a transition from state 0 to state 1 with the symbol x
        # then a[0][1]=x
        self.A = {}
        # B[n] holds the regular expression that describes how a final state
        # can be reached from state n
        self.B = {}
        self.epsilon = ''
        self.empty = None

    def _bfs_sort(self, start):
        """
        maintain a map of states distance using BFS
        Args:
            start (fst state): The initial DFA state
        Returns:
            list: An ordered list of DFA states
                  using path distance
        """
        pathstates = {}
        # maintain a queue of nodes to be visited. Both current and previous
        # node must be included.
        queue = []
        # push the first path into the queue
        queue.append([0, start])
        pathstates[start.stateid] = 0
        while queue:
            # get the first node from the queue
            leaf = queue.pop(0)
            node = leaf[1]
            pathlen = leaf[0]
            # enumerate all adjacent nodes, construct a new path and push it
            # into the queue
            for arc in node.arcs:
                next_state = self.mma[arc.nextstate]
                if next_state.stateid not in pathstates:
                    queue.append([pathlen + 1, next_state])
                    pathstates[next_state.stateid] = pathlen + 1
        orderedstatesdict = OrderedDict(
            sorted(
                pathstates.items(),
                key=lambda x: x[1],
                reverse=False))
        for state in self.mma.states:
            orderedstatesdict[state.stateid] = state
        orderedstates = [x[1] for x in list(orderedstatesdict.items())]
        return orderedstates

    def star(self, input_string):
        """
        Kleene star operation
        Args:
            input_string (str): The string that the kleene star will be made
        Returns:
            str: The applied Kleene star operation on the input string
        """

        if input_string != self.epsilon and input_string != self.empty:
            return "(" + input_string + ")*"
        else:
            return ""

    def _create_map(self):
        """Initialize Brzozowski Algebraic Method"""

        # at state i is represented by the regex self.B[i]
        for state_a in self.mma.states:
            self.A[state_a.stateid] = {}
            # Create a map state to state, with the transition symbols
            for arc in state_a.arcs:
                if arc.nextstate in self.A[state_a.stateid]:
                    self.A[state_a.stateid][arc.nextstate].append(self.mma.isyms.find(arc.ilabel))
                else:
                    self.A[state_a.stateid][arc.nextstate] = [self.mma.isyms.find(arc.ilabel)]
            if state_a.final:
                self.A[state_a.stateid]['string'] = ['']

       # In both self.B and Self.A there are empty spots that we will fill in _brzozowski_algebraic_method_solve

    def _fix_pattern(self):
        for state_a in self.mma.states:
            for target in self.A[state_a.stateid]:
                if len(self.A[state_a.stateid][target]) == 0:
                    self.A[state_a.stateid][target] = ''
                elif len(self.A[state_a.stateid][target]) == 1:
                    self.A[state_a.stateid][target] = self.A[state_a.stateid][target][0]
                elif len(self.A[state_a.stateid][target]) == len(self.alphabet):
                    self.A[state_a.stateid][target] = '.'
                elif len([a for a in self.A[state_a.stateid][target] if len(a) > 1]) == 0 and len(self.A[state_a.stateid][target]) > len(self.alphabet) / 2:
                    self.A[state_a.stateid][target] = '[^' + ''.join(
                        filter(lambda x: x not in self.A[state_a.stateid][target], self.alphabet)) + ']'
                elif len([a for a in self.A[state_a.stateid][target] if len(a) > 1]) == 0 and len(self.A[state_a.stateid][target]) < len(self.alphabet) / 2:
                    self.A[state_a.stateid][target] = '[' + ''.join(self.A[state_a.stateid][target]) + ']'
                elif len([a for a in self.A[state_a.stateid][target] if len(a) > 1]) > 0:
                    self.A[state_a.stateid][target] = '(' + '|'.join(self.A[state_a.stateid][target]) + ')'

    def partitionAlphabet(self, alphabetLen, parsedInput):
        intoInput = []
        notInInput = []
        for c in self.alphabet:
            if c in parsedInput:
                intoInput.append(c)
            else:
                notInInput.append(c)
        if len(intoInput) == alphabetLen:
            return '.'
        if (len(intoInput) < len(notInInput)):
            final = '[' + ''.join(intoInput) + ']'
        else:
            final = '[^' + ''.join(notInInput) + ']'
        return final

    def replaceAlphabet(self, input):
        if "|" not in input:
            return input
        alphabetLen = len(self.alphabet)
        parsedInput = input.split("|")
        if "|||" in input:
            parsedInput.append("|")

        for c in parsedInput:
            if len(c) > 1:
                return input

        final = self.partitionAlphabet(alphabetLen, parsedInput)
        return final

    def cleaner(self, res):
        res = self.replaceAlphabet(res)
        import re
        if '|' in re.sub('\(.*\)', '', res):
            res = '(' + res + ')'
        return res

    def cleanerS(self, res):
        res = self.replaceAlphabet(res)
        import re
        if '|' in re.sub(r'\([^)]*\)', '', res):
            res = '(' + res + ')*'
        return res

    def findWhichSymbolsFromOtherStatesConnectToMe(self, state):
        array = []
        for n in self.A:
            if n != state and state in self.A[n]:
                if self.A[n][state] not in array:
                    array.append(self.A[n][state])
        return array

    def checkIfSameStateNotNeeded(self, state, same):
        array = self.findWhichSymbolsFromOtherStatesConnectToMe(state)
        parts = []
        if len(array) == 0:
            return same
        try:
            if len(self.A[state]['same']) > 3 and self.A[state]['same'][0] == '(' and self.A[state]['same'][
                                                                                            len(self.A[state][
                                                                                                    'same']) - 1:len(
                                                                                                self.A[state][
                                                                                                        'same']) + 1]:
                parts.append(self.A[state]['same'][1:len(self.A[state]['same']) - 2])
            else:
                parts = re.findall(r'\((.+?)\)\*', self.A[state]['same'])

            for n in parts:
                if n[len(n) - 1] != array[0]:
                    return same
            return ''
        except AttributeError:
            return same

    def checkIfBothAreNeeded(self, inputA, inputB):  # based on the fact that ([^b]*|b([^i])*) is equal to  .*
        try:
            input1 = re.search('\(\[\^(.+?)\]\)\*', inputA).group(1)
        except AttributeError:
            input1 = inputA
        try:
            input2FirstChar = re.search('\((.+?)\)\*', inputB).group(1)[0]
        except AttributeError:
            input2FirstChar = inputB[0]

        if len(input1) == 1 and input1 == input2FirstChar:
            return 1
        else:
            return -1

    def replaceAlphabetMixed(self, input1, input2):
        found1 = 0
        found2 = 0
        if "|" not in input1:
            final1 = input1
            if len(input1) > 1:
                found1 = 1
        else:
            parsedInput = input1.split("|")
            for c in parsedInput:
                if len(c) > 1:
                    final1 = input1
                    found1 = 1
                    break

        if "|" not in input2:
            final2 = input2
            if len(input2) > 1:
                found2 = 1
        else:
            parsedInput = input2.split("|")
            for c in parsedInput:
                if len(c) > 1:
                    final2 = input2
                    found2 = 1
                    break

        if found1 == 1 and found2 == 0:
            if (self.checkIfBothAreNeeded(input1, input2) == 1):
                return '(.)*'
            final2 = self.replaceAlphabet(input2)
        if found1 == 0 and found2 == 1:
            if (self.checkIfBothAreNeeded(input1, input2) == 1):
                return '(.)*'
            final1 = self.replaceAlphabet(input1)
        if found1 == 1 and found2 == 1:
            if (self.checkIfBothAreNeeded(input1, input2) == 1):
                return '(.)*'
        if found1 == 0 and found2 == 0:
            final1 = input1
            final2 = input2
        return final1 + "|" + final2

    def getMaxInternalPath(self, inkeys):
        maxkey = 0
        keys = inkeys
        if 'same' in keys:
            keys.remove('same')
        if 'rest' in keys:
            keys.remove('rest')
        if 'string' in keys:
            keys.remove('string')
        if len(keys) > 0:
            maxkey = sorted(keys, reverse=True)[0]
        return maxkey

    def existsPath(self, initial, finalstate, visited):
        for i in self.A:
            if finalstate in self.A[i] and i not in visited:
                if i == initial:
                    return 1, self.A[i][finalstate], finalstate
                else:
                    visited.append(i)
                    res, key, transfer_state = self.existsPath( initial, i, visited)
                    if res == 1:
                        return 1, key, finalstate
                    else:
                        continue
            else:
                continue
        return 0, 0, 0
    def fixBrzozowskiBackwardLoopRemoval(self):

        # Remove targets that point to start
        for i in self.A:
            if i != 0:
                for target in self.A[i].keys():
                    if target == 0:
                        self.A[i].pop(target)

        for i in self.A:
            poplist = []
            for target in self.A[i]:
                if target < i:
                    exists, transfer_value, transfer_state = self.existsPath(target, i, [])

                    if exists and (transfer_value == self.A[i][target] or (target in self.A[target] and (self.A[target][target] == self.A[i][target] or len(
                                set(self.A[i][target]) - set(self.A[target][target])) == 0))):
                        poplist.append(target)
            for k in poplist:
                self.A[i].pop(k)


    def _Lexical(self):
        found = 0
        start = 0
        # First Try to Optimize the Strings by doing the possible replacements of special cases to string. e.c change a loop with character e to (e)*
        for state_a in self.A.keys():
            # Loopholes
            if state_a in self.A[state_a]:
                same = ''
                if 'same' in self.A[state_a]:
                    same = '|' + self.A[state_a]['same']
                    self.A[state_a]['same'] = '(' + ''.join(self.replaceAlphabet(self.A[state_a][state_a])) + ')*' + same
                del(self.A[state_a][state_a])
            # Final State
            if len(self.A[state_a].keys()) == 1 and 'string' in self.A[state_a].keys():
                for state_b in self.A:
                    if state_b != state_a  and state_b in self.A:
                        if state_a in self.A[state_b]:
                            if 'rest' in self.A[state_b] and self.A[state_b]['rest'] != '' and  self.A[state_b]['rest'] == self.cleaner(self.A[state_a]['string']):


                                if 'string' not in self.A[state_b]:
                                    self.A[state_b]['string'] = "(" + self.A[state_b][state_a] + ")?"+self.A[state_b]['rest']
                                else:
                                    self.A[state_b]['string'] = "("+ self.A[state_b]['string']+"|(" + self.A[state_b][state_a] + ")?"+self.A[state_b]['rest']+")"
                                del (self.A[state_b][state_a])
                                del (self.A[state_b]['rest'])
                                found = 1
                            else:
                                newinput = ''.join(self.A[state_b][state_a]) + self.cleaner(self.A[state_a]['string'])
                                if 'rest' in self.A[state_b] and self.A[state_b]['rest'] != '':
                                    newinput = '(' + newinput + '|' + self.A[state_b]['rest'] + ')'

                                self.A[state_b]['rest'] = newinput
                                del (self.A[state_b][state_a])
                                found = 1
                if state_a != start:
                    del (self.A[state_a])
            if state_a != start and state_a in self.A and len(self.A[state_a].keys()) == 1 and 'same' in self.A[state_a].keys():
                del (self.A[state_a])
                for state_b in self.A:
                    if state_b in self.A and state_a in self.A[state_b]:
                        del (self.A[state_b][state_a])
                        found = 1
            # Clean Loophole
            if state_a in self.A and len(self.A[state_a].keys()) == 1 and 'same' in self.A[state_a].keys():
                del (self.A[state_a])
                found = 1
            # Final State
            if state_a in self.A and len(self.A[state_a].keys()) == 1 and 'rest' in self.A[state_a]:
                self.A[state_a]['string'] = self.A[state_a]['rest']
                del (self.A[state_a]['rest'])
                found = 1
            # Final State with alternative path
            if state_a in self.A and len(self.A[state_a].keys()) == 2 and 'string' in self.A[state_a] and 'rest' in \
                    self.A[state_a]:
                if self.A[state_a]['rest'] != '':
                    self.A[state_a]['string'] = self.cleaner(self.A[state_a]['string']) + '(' + self.replaceAlphabet(
                        self.A[state_a]['rest']) + ')?'
                del (self.A[state_a]['rest'])
                found = 1
            # State with alternative path and self loop
            if state_a in self.A and len(self.A[state_a].keys()) == 2 and 'same' in self.A[state_a] and 'rest' in \
                    self.A[state_a]:
                self.A[state_a]['rest'] = self.checkIfSameStateNotNeeded( state_a, self.cleanerS(self.A[state_a][
                    'same'])) + self.cleaner(self.A[state_a]['rest'])
                del (self.A[state_a]['same'])
                found = 1
            # Final State with self loop
            if state_a in self.A and len(self.A[state_a].keys()) == 2 and 'string' in self.A[state_a] and 'same' in \
                    self.A[state_a]:
                self.A[state_a]['string'] = self.checkIfSameStateNotNeeded( state_a, self.cleanerS(self.A[state_a]['same'])) + self.cleaner(self.A[state_a]['string'])
                del (self.A[state_a]['same'])
                found = 1
            # Final State with alternative path and self loop
            if state_a in self.A and len(self.A[state_a].keys()) == 3 and 'string' in self.A[state_a] and 'rest' in \
                    self.A[state_a] and 'same' in self.A[state_a]:
                if self.A[state_a]['rest'] != '':
                    add = '(' + self.A[state_a]['rest'] + ')?'
                self.A[state_a]['string'] = self.checkIfSameStateNotNeeded(state_a, self.cleanerS(self.A[state_a]['same'])) + self.cleaner(self.A[state_a]['string']) + add
                del (self.A[state_a]['rest'])
                del (self.A[state_a]['same'])
                found = 1
            # No path State
            if state_a != start and state_a in self.A and len(self.A[state_a].keys()) == 0:
                del (self.A[state_a])
                for state_b in self.A:
                    if state_b in self.A and state_a in self.A[state_b]:
                        del (self.A[state_b][state_a])

        return found

    def fixBrzozowskiAdvanced(self, found):
        start = 0
        for state_a in self.A:
            maxkey = self.getMaxInternalPath(self.A[state_a].keys())
            if state_a != start and state_a in self.A and maxkey < state_a and (
                (len(self.A[state_a].keys()) > 1) or (len(self.A[state_a].keys()) == 1 and 'same' not in self.A[state_a])):
                same = ''
                if 'same' in self.A[state_a]:
                    same = self.checkIfSameStateNotNeeded( state_a, self.cleanerS(self.A[state_a]['same']))
                # del(self.A[i]['same'])
                for state_b in [x.stateid for x in self._bfs_sort(sorted(self.mma.states,key=attrgetter('initial'),reverse=True)[0])]:
                    if state_b != state_a and state_b in self.A:
                        if state_a in self.A[state_b]:
                            temp = self.cleaner(self.A[state_b][state_a])
                            del (self.A[state_b][state_a])
                            for newTarget in self.A[state_a]:
                                if newTarget != 'same':
                                    transValue = same + self.cleaner( self.A[state_a][newTarget])
                                    found = 1
                                    add = temp + transValue
                                    if newTarget == state_b:
                                        newTarget = 'same'
                                        add = '(' + self.replaceAlphabet(add) + ')*'
                                    if ((newTarget not in self.A[state_b]) or (self.A[state_b][newTarget] == '')):
                                        self.A[state_b][newTarget] = add
                                    else:
                                        self.A[state_b][newTarget] = self.replaceAlphabetMixed(self.A[state_b][newTarget], add)
                                    found = 1

                            if found == 1:
                                break

        return found


    def _solve(self):
        self._create_map()
        self._fix_pattern()
        self.fixBrzozowskiBackwardLoopRemoval()  # Remove backward paths because of kleene star

        while 1:
            found = self._Lexical()
            if found == 0:
                found = self.fixBrzozowskiAdvanced(found)
            if found == 0:
                break
        if 'string' in self.A[0]:
            return ''.join(self.A[0]['string'].splitlines())

    def get_regex(self):
        """Generate regular expressions from DFA automaton"""

        return self._solve()

def main():
    """
    Testing function for DFA brzozowski algebraic method Operation
    """
    argv = sys.argv
    if len(argv) < 2:
        targetfile = 'target.y'
    else:
        targetfile = argv[1]
    print 'Parsing ruleset: ' + targetfile,
    flex_a = Flexparser()
    mma = flex_a.yyparse(targetfile)
    print 'OK'
    print 'Perform minimization on initial automaton:',
    print 'OK'
    print 'Perform Brzozowski on minimal automaton:',
    brzozowski_a = Regex(mma)
    mma_regex = brzozowski_a.get_regex()
    print mma_regex

if __name__ == '__main__':
    main()



"""This module generates a PDA using a CFG as input"""
from FAdo import cfg
from cnfpda import CnfPda

class CfgPDA():
    """Create a PDA from a CFG"""

    def __init__(self, alphabet=None):
        """
        Initialize values
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        self.alphabet = alphabet

    def _extract_alphabet(self, grammar):
        """
        Extract an alphabet from the given grammar.
        """
        alphabet = set([])
        for terminal in grammar.Terminals:
            alphabet |= set([x for x in terminal])
        self.alphabet = list(alphabet)

    def _read_file(self, fname):
        """
        Read file containing CFG
        Args:
            fname (str): The file name
        Returns:
            list: The list of CFG rules
        """
        with open(fname) as input_file:
            re_grammar = [x.strip('\n') for x in input_file.readlines()]
        return re_grammar

    def _mpda(self, re_grammar, splitstring=0):
        """
        Args:
            re_grammar (list): A list of grammar rules
            splitstring (bool): A boolean for enabling or disabling
                            the splitting of symbols using a space
        Returns:
            PDA: The generated PDA
        """
        grammar = cfg.gRules(re_grammar, ":")
        cnfgrammar = cfg.CNF(grammar)

        if not self.alphabet:
            self._extract_alphabet(cnfgrammar)

        cnftopda = CnfPda(self.alphabet)
        productions = {}
        nonterminals = []
        nonterminals.append(cnfgrammar.Start)
        for key in list(cnfgrammar.Nonterminals):
            if key != cnfgrammar.Start:
                nonterminals.append(key)
        for key in list(cnfgrammar.Nonterminals):
            j = 0
            productions[key] = {}
            # print 'testing '+key
            for pair in cnfgrammar.Rules:
                cnf_form = list(pair)
                if cnf_form[0] == key:
                    productions[key][j] = {}
                    if isinstance(cnf_form[1], type(())):
                        #            print list(p[1])
                        productions[key][j]['b0'] = list(cnf_form[1])[0]
                        productions[key][j]['b1'] = list(cnf_form[1])[1]
                    else:
                        #           print p[1]
                        productions[key][j]['a'] = cnf_form[1]
                    j = j + 1
        return cnftopda.initialize(
            nonterminals, productions, list(
                cnfgrammar.Terminals), splitstring)

    def yyparse(self, cfgfile, splitstring=0):
        """
        Args:
            cfgfile (str): The path for the file containing the CFG rules
            splitstring (bool): A boolean for enabling or disabling
                            the splitting of symbols using a space
        Returns:
            PDA: The generated PDA
        """
        re_grammar = self._read_file(cfgfile)
        mma = self._mpda(re_grammar, splitstring)
        return mma



"""
This module performs all basic DFA operations.
It is an interface for pyfst.
"""
# /usr/bin/python

from operator import attrgetter
import fst
from alphabet import createalphabet

EPSILON = fst.EPSILON

def TropicalWeight(param):
    """
    Returns fst TropicalWeight
    Args:
        param (str): The input
    Returns:
        bool: The arc weight
    """
    return fst.TropicalWeight(param)

class FstDFA(fst.StdAcceptor):
    """
    Contains extra method to consume input and produce outputs.
    The underline library is pyfst, the python bindings of openFST library.
    """

    def __init__(self, alphabet = createalphabet()):
        """
        Args:
            alphabet (list): pyfst input symbol list
        Returns:
            None
        """
        isyms = None
        self.alphabet = alphabet
        fst.StdAcceptor.__init__(self, isyms)
        num = 1
        for char in self.alphabet:
            self.isyms.__setitem__(char, num)
            num = num + 1

    def fixminimized(self, alphabet):
        """
        After pyfst minimization,
        all unused arcs are removed,
        and all sink states are removed.
        However this may break compatibility.
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        endstate = len(list(self.states))
        for state in self.states:
            for char in alphabet:
                found = 0
                for arc in state.arcs:
                    if self.isyms.find(arc.ilabel) == char:
                        found = 1
                        break
                if found == 0:
                    self.add_arc(state.stateid, endstate, char)

        self[endstate].final = TropicalWeight(float('inf'))

        for char in alphabet:
            self.add_arc(endstate, endstate, char)

    def _addsink(self, alphabet):
        """
        Adds a sink state
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        endstate = len(list(self.states))
        for state in self.states:
            for char in alphabet:
                found = 0
                for arc in state.arcs:
                    if self.isyms.find(arc.ilabel) == char:
                        found = 1
                        break
                if found == 0:
                    self.add_arc(state.stateid, endstate, char)

        self[endstate].final = TropicalWeight(float('inf'))

        for char in alphabet:
            self.add_arc(endstate, endstate, char)

    def _path_to_str(self, path):
        """
        Convert a path to the string representing the path
        Args:
            path (tuple): A tuple of arcs
        Returns:
            inp (str): The path concatenated as as string
        """
        inp = ''
        for arc in path:
            i = self.isyms.find(arc.ilabel)
            # Ignore \epsilon transitions both on input
            if i != fst.EPSILON:
                inp += i
        return inp

    def init_from_acceptor(self, acceptor):
        """
        Adds a sink state
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        states = sorted(
            acceptor.states,
            key=attrgetter('initial'),
            reverse=True)
        for state in states:
            for arc in state.arcs:
                itext = acceptor.isyms.find(arc.ilabel)
                if itext in self.alphabet:
                    self.add_arc(state.stateid, arc.nextstate, itext)
            if state.final:
                self[state.stateid].final = True
            if state.initial:
                self[state.stateid].initial = True

    def consume_input(self, inp):
        """
        Return True/False if the machine accepts/reject the input.
        Args:
            inp (str): input string to be consumed
        Returns:
            bool: A true or false value depending on if the DFA
                accepts the provided input
        """
        cur_state = sorted(
            self.states,
            key=attrgetter('initial'),
            reverse=True)[0]
        while len(inp) > 0:
            found = False
            for arc in cur_state.arcs:
                if self.isyms.find(arc.ilabel) == inp[0]:
                    cur_state = self[arc.nextstate]
                    inp = inp[1:]
                    found = True
                    break
            if not found:
                return False
        return cur_state.final != TropicalWeight(float('inf'))

    def empty(self):
        """""
        Return True if the DFA accepts the empty language.
        """
        return len(list(self.states)) == 0

    def random_strings(self, string_length=1):
        """
        Generate string_length random strings that belong to the automaton.
        Args:
            string_length (integer): The size of the random string
        Returns:
            str: The generated string
        """
        str_list = []
        for path in self.uniform_generate(string_length):
            str_list.append(self._path_to_str(path))
        return str_list

    def complement(self, alphabet):
        """
        Generate the complement of a DFA automaton
        Args:
            alphabet (list): The input alphabet
        Returns:
            None
        """
        self._addsink(alphabet)
        states = sorted(self.states, key=attrgetter('initial'), reverse=True)
        for state in states:
            if state.final:
                state.final = False
            else:
                state.final = True

    def save(self, txt_fst_filename):
        """
        Save the machine in the openFST format in the file denoted by
        txt_fst_filename.
        Args:
            txt_fst_filename (str): The name of the file
        Returns:
            None
        """
        txt_fst = open(txt_fst_filename, 'w+')
        states = sorted(self.states, key=attrgetter('initial'), reverse=True)
        for state in states:
            for arc in state.arcs:
                itext = self.isyms.find(arc.ilabel)
                otext = self.osyms.find(arc.ilabel)
                txt_fst.write(
                    '{}\t{}\t{}\t{}\n'.format(
                        state.stateid,
                        arc.nextstate,
                        itext.encode('hex'),
                        otext.encode('hex')))
            if state.final:
                txt_fst.write('{}\n'.format(state.stateid))
        txt_fst.close()

    def load(self, txt_fst_filename):
        """
        Save the transducer in the text file format of OpenFST.
        The format is specified as follows:
            arc format: src dest ilabel olabel [weight]
            final state format: state [weight]
        lines may occur in any order except initial state must be first line
        Args:
            txt_fst_filename (string): The name of the file
        Returns:
            None
        """
        with open(txt_fst_filename, 'r') as txt_fst:
            for line in txt_fst:
                line = line.strip()
                splitted_line = line.split()
                if len(splitted_line) == 1:
                    self[int(splitted_line[0])].final = True
                else:
                    self.add_arc(int(splitted_line[0]), int(
                        splitted_line[1]), splitted_line[2].decode('hex'))




"""
This module transforms a file containing regular expressions for flex parser
into a DFA object (pyfst automaton)
"""
import os
import random
import re
import string
import tempfile
from subprocess import call
from sys import argv

from alphabet import createalphabet
from dfa import DFA


class Flexparser:
    """
    This class parses compiles a file containing regular
    expressions into a flex compiled file, and then parses
    the flex DFA into a pyfst DFA
    """

    outfile = ""

    def __init__(self, alphabet=None):
        """
        Initialization function
        Args:
            alphabet (list): input alphabet
        Returns:
            None
        """
        if alphabet is not None:
            self.alphabet = alphabet
        else:
            self.alphabet = []

    def _create_automaton_from_regex(self, myfile):
        """
        Generates a flex compiled file using myfile
        as input
        Args:
            myfile (str): Flex file to be parsed
        Returns:
            None
        """
        call(["flex", "-o", self.outfile, "-f", myfile])

    def _read_transitions(self):
        """
        Read DFA transitions from flex compiled file
        Args:
            None
        Returns:
            list: The list of states and the destination for a character
        """
        states = []
        i = 0
        regex = re.compile('[ \t\n\r:,]+')
        found = 0  # For maintaining the state of yy_nxt declaration
        state = 0  # For maintaining the state of opening and closing tag of yy_nxt
        substate = 0  # For maintaining the state of opening and closing tag of each set in yy_nxt
        mapping = []  # For writing each set of yy_next
        cur_line = None
        with open(self.outfile) as flex_file:
            for cur_line in flex_file:
                if cur_line[0:35] == "static yyconst flex_int16_t yy_nxt[":
                    found = 1
                    # print 'Found yy_next declaration'
                    continue
                if found == 1:
                    if state == 0 and cur_line[0:5] == "    {":
                        state = 1
                        continue
                    if state == 1 and cur_line[0:7] == "    } ;":
                        state = 0
                        break

                    if substate == 0 and cur_line[0:5] == "    {":
                        mapping = []
                        substate = 1
                        continue
                    if substate == 1:
                        if cur_line[0:6] != "    },":
                            cur_line = "".join(cur_line.split())
                            if cur_line == '':
                                continue
                            if cur_line[cur_line.__len__() - 1] == ',':
                                splitted_line = regex.split(
                                    cur_line[:cur_line.__len__() - 1])
                            else:
                                splitted_line = regex.split(cur_line)
                            mapping = mapping + splitted_line
                            continue
                        else:
                            cleared = []
                            for j in mapping:
                                cleared.append(int(j))
                            states.append(cleared)
                            mapping = []
                            substate = 0

        return states

    def _read_accept_states(self):
        """
        Read DFA accepted states from flex compiled file
        Args:
            None
        Returns:
            list: The list of accepted states
        """
        states = []
        i = 0
        regex = re.compile('[ \t\n\r:,]+')
        found = 0  # For maintaining the state of yy_accept declaration
        state = 0  # For maintaining the state of opening and closing tag of yy_accept
        mapping = [] # For writing each set of yy_accept
        cur_line = None
        with open(self.outfile) as flex_file:
            for cur_line in flex_file:
                if cur_line[0:37] == "static yyconst flex_int16_t yy_accept":
                    found = 1
                    # print 'Found yy_accept declaration'
                    continue
                if found == 1:
                    # print x
                    if state == 0 and cur_line[0:5] == "    {":
                        mapping.append(0)  # there is always a zero there
                        state = 1
                        continue

                    if state == 1:
                        if cur_line[0:7] != "    } ;":
                            cur_line = "".join(cur_line.split())
                            if cur_line == '':
                                continue
                            if cur_line[cur_line.__len__() - 1] == ',':
                                splitted_line = regex.split(
                                    cur_line[:cur_line.__len__() - 1])
                            else:
                                splitted_line = regex.split(cur_line)
                            mapping = mapping + splitted_line
                            continue
                        else:
                            cleared = []
                            for j in mapping:
                                cleared.append(int(j))
                            max_value = max(cleared)
                            for i in range(0, len(cleared)):
                                if cleared[i] > 0 and cleared[
                                        i] < (max_value - 1):
                                    states.append(i)
                            return states
        return []

    def _read_null_transitions(self):
        """
        Read DFA'input_string NULL transitions from flex compiled file
        Args:
            None
        Returns:
            list: The list of state transitions for no character
        """
        states = []
        regex = re.compile('[ \t\n\r:,]+')
        found = 0  # For maintaining the state of yy_NUL_trans declaration
        state = 0  # For maintaining the state of opening and closing tag of yy_NUL_trans
        mapping = []  # For writing each set of yy_NUL_trans
        cur_line = None
        with open(self.outfile) as flex_file:
            for cur_line in flex_file:
                if cur_line[0:len("static yyconst yy_state_type yy_NUL_trans")
                            ] == "static yyconst yy_state_type yy_NUL_trans":
                    found = 1
                    # print 'Found yy_next declaration'
                    continue
                if found == 1:
                    if state == 0 and cur_line[0:5] == "    {":
                        mapping.append(0)  # there is always a zero there
                        state = 1
                        continue
                    if state == 1:
                        if cur_line[0:7] != "    } ;":
                            cur_line = "".join(cur_line.split())
                            if cur_line == '':
                                continue
                            if cur_line[cur_line.__len__() - 1] == ',':
                                splitted_line = regex.split(
                                    cur_line[:cur_line.__len__() - 1])
                            else:
                                splitted_line = regex.split(cur_line)
                                #  print y
                            mapping = mapping + splitted_line
                            continue
                        else:
                            cleared = []
                            for j in mapping:
                                cleared.append(int(j))
                            states = cleared
                            mapping = []
                            state = 0

        return states

    def _create_states(self, states_num):
        """
        Args:
            states_num (int): Number of States
        Returns:
            list: An initialized list
        """
        states = []
        for i in range(0, states_num):
            states.append(i)
        return states

    def _add_sink_state(self, states):
        """
        This function adds a sing state in the total states
        Args:
            states (list): The current states
        Returns:
            None
        """
        cleared = []
        for i in range(0, 128):
            cleared.append(-1)
        states.append(cleared)

    def _create_delta(self):
        """
        This function creates the delta transition
        Args:
            startState (int): Initial state of automaton
        Results:
            int, func: A number indicating the total states, and the delta function
        """
        states = self._read_transitions()
        total_states = len(states)
        self._add_sink_state(states)
        nulltrans = self._read_null_transitions()

        def delta(current_state, character):
            """
            Sub function describing the transitions
            Args:
                current_state (str): The current state
                character (str): The input character
            Returns:
                str: The next state
            """
            if character != '':
                newstate = states[current_state][ord(character)]
                if newstate > 0:
                    return newstate
                else:
                    return total_states
            else:
                return nulltrans[current_state]

        return total_states + 1, delta

    def yyparse(self, lexfile):
        """
        Args:
            lexfile (str): Flex file to be parsed
        Returns:
            DFA: A dfa automaton
        """
        temp = tempfile.gettempdir()
        self.outfile = temp+'/'+''.join(
            random.choice(
                string.ascii_uppercase + string.digits) for _ in range(5)) + '_lex.yy.c'
        self._create_automaton_from_regex(lexfile)
        states_num, delta = self._create_delta()
        states = self._create_states(states_num)
        accepted_states = self._read_accept_states()
        if self.alphabet != []:
            alphabet = self.alphabet
        else:
            alphabet = createalphabet()
        mma = DFA(alphabet)
        for state in states:
            if state != 0:
                for char in alphabet:
                    nextstate = delta(state, char)
                    mma.add_arc(state - 1, nextstate - 1, char)
                if state in accepted_states:
                    mma[state - 1].final = True
        if os.path.exists(self.outfile):
            os.remove(self.outfile)
        return mma


def main():
    """
    Testing function for Flex Regular Expressions to FST DFA
    """
    if len(argv) < 2:
        print 'Usage: %s fst_file [optional: save_file]' % argv[0]
        return
    flex_a = Flexparser()
    mma = flex_a.yyparse(argv[1])
    mma.minimize()
    print mma
    if len(argv) == 3:
        mma.save(argv[2])


if __name__ == '__main__':
    main()



"""
This module performs Diff operation
between a PDA and a DFA
"""
import time
from sys import argv

import datetime
import dateutil.relativedelta
from FAdo import cfg
from alphabet import createalphabet
from cfggenerator import CFGGenerator
from cfgpda import CfgPDA
from flex2fst import Flexparser
from pda import PDA, PDAState
from pdacnf import PdaCnf, IntersectionHandling, SimplifyStateIDs, ReadReplace, ReducePDA
from pdastring import PdaString
from dfa import TropicalWeight

class PdaDiff():
    """Performs Diff operation between a PDA and a DFA"""

    def __init__(self, input_pda_a, input_dfa_b, alphabet):
        """
        Args:
            input_pda_a (PDA): The input PDA
            input_dfa_b (DFA): The input pyfst DFA
        Returns:
            None
        """
        self.mma = input_pda_a
        self.mmb = None
        if input_dfa_b:
            self.mmb = input_dfa_b.copy()
        self.mmc = None
        self.alphabet = alphabet

    def _delta(self, graph, cur_state, char):
        """
        Args:
            graph (Fst Acceptor): The DFA
            cur_state (Fst State): The current State
            char (Char): The input character
        Returns:
            (Fst State): The destination state
        """
        for arc in cur_state.arcs:
            if graph.isyms.find(arc.ilabel) == char:
                return graph[arc.nextstate]
        return None

    def _break_terms(self):
        counter = len(self.mma.s)
        for state in self.mma.s.keys():
            if self.mma.s[state].type == 3:
                for transition in self.mma.s[state].trans.keys():
                    for transition in self.mma.s[state].trans.keys():
                        for record in self.mma.s[state].trans[transition]:
                            if len(record) > 1:
                                    value = record
                                    i = self.mma.s[state].trans[transition].index(record)
                                    del self.mma.s[state].trans[transition][i]
                                    if len(self.mma.s[state].trans[transition]) == 0:
                                        del self.mma.s[state].trans[transition]
                                    tempstate = PDAState()
                                    tempstate.id = counter
                                    tempstate.sym = 0
                                    tempstate.type = 3
                                    tempstate.trans = {}
                                    tempstate.trans[transition] = [value[-1]]
                                    self.mma.s[counter] = tempstate
                                    counter = counter + 1
                                    for character in reversed(value[1:-1]):
                                        tempstate = PDAState()
                                        tempstate.id = counter
                                        tempstate.sym = 0
                                        tempstate.type = 3
                                        tempstate.trans = {}
                                        tempstate.trans[counter-1] = [character]
                                        self.mma.s[counter] = tempstate
                                        counter = counter + 1

                                    self.mma.s[state].trans[counter-1] = [value[0]]
        self.mma.n = counter-1


    def _intesect(self):
        """The intesection of a PDA and a DFA"""
        p1automaton = self.mma
        p2automaton = self.mmb
        p3automaton = PDA(self.alphabet)
        self._break_terms()
        p1counter = 0
        p3counter = 0
        p2states = list(p2automaton.states)
        print 'PDA States: ' + repr(p1automaton.n)
        print 'DFA States: ' + repr(len(list(p2states)))
        ignorechars = p1automaton.nonterminals+ [0] + ['@closing']
        del(ignorechars[ignorechars.index('S')])
        while p1counter < p1automaton.n + 1:
            p1state = p1automaton.s[p1counter]
            p2counter = 0
            while p2counter < len(list(p2states)):
                p2state = p2states[p2counter]
                tempstate = PDAState()
                tempstate.id = (p1state.id, p2state.stateid)
                tempstate.sym = p1state.sym
                tempstate.type = p1state.type
                tempstate.trans = {}
                found = 0
                for char in self.alphabet:
                    if char in ignorechars:
                        continue
                    # DFA has single destination from a state
                    p2dest = self._delta(p2automaton, p2state, char)
                    # PDA may have multiple destinations from a state
                    # print p1state.trans
                    if p2dest is not None:
                        for potential in p1state.trans:
                            if char in p1state.trans[potential]:
                                found = 1
                                p1dest = potential
                                if (p1dest,
                                        p2dest.stateid) not in tempstate.trans:
                                    tempstate.trans[
                                        (p1dest, p2dest.stateid)] = []
                                    # print 'Appending A Transition to
                                    # ('+`p1dest`+','+`p2dest.stateid`+') for
                                    # input '+`char`
                                tempstate.trans[
                                    (p1dest, p2dest.stateid)].append(char)

                                # THEN THE NONTERMINALS + 0 3 transitions
                                # print p1state.trans
                                #               print p1automaton.nonterminals
                if found == 0 and  p1state.type == 3 and len(p1state.trans) >0:
                    assert 1==1,'Check Failed: A READ state with transitions' \
                                ' did not participate in the cross product'
                if p2dest is not None:
                    for nonterm in p1automaton.nonterminals + \
                            [0] + ['@closing']:
                        for potential in p1state.trans:
                            if nonterm in p1state.trans[potential]:
                                p1dest = potential
                                if (p1dest,
                                        p2state.stateid) not in tempstate.trans:
                                    tempstate.trans[
                                        (p1dest, p2state.stateid)] = []
                                # print 'Appending B Transition to
                                # ('+`p1dest`+','+`p2state.stateid`+') for
                                # input '+`nonterm`
                                tempstate.trans[
                                    (p1dest, p2state.stateid)].append(nonterm)
                p3automaton.s[p3counter] = tempstate
                p3counter = p3counter + 1
                p2counter = p2counter + 1
            p1counter = p1counter + 1
        # print 'Total States Appended '+`len(p3automaton.input_string)`
        p3automaton.n = p3counter - 1

        p3automaton.accepted = []
        for state in p2automaton.states:
            if state.final != TropicalWeight(float('inf')):
                p3automaton.accepted.append(state.stateid)
        return p3automaton

    def diff(self):
        """The Difference between a PDA and a DFA"""
        self.mmb.complement(self.alphabet)
        self.mmb.minimize()
        print 'start intersection'
        self.mmc = self._intesect()
        print 'end intersection'
        return self.mmc

    def get_string(self):
        """
        Returns a string from the Diff resutl.
        Depending on the method, either the string will
        be generated directly from the PDA using the state
        removal method, or the PDA will be first translated to
        a CFG and then a string will be generated from the CFG
        Args:
             None
        Returns:
            A string from the Diff
        """
        return_string = None
        if not self.mmc:
            return ""
        method = 'PDASTRING'
        if method == 'PDASTRING':
            stringgen = PdaString()
            print '* Reduce PDA using DFA BFS (remove unreachable states):'
            newpda = self.mmc.s
            handle = IntersectionHandling()
            newpda = handle.get(newpda, self.mmc.accepted)
            reduce_b = ReducePDA()
            newpda = reduce_b.get(newpda)
            #simply = SimplifyStateIDs()
            #newpda, biggestid, newaccepted = simply.get(
            #    newpda, self.mmc.accepted)
            print "- Total PDA states after reduction are " + repr(len(newpda))
            return_string = stringgen.init(newpda, self.mmc.accepted)
            if return_string is not None:
                return_string = return_string[0]
        elif method == 'PDACFGSTRING':

            optimized = 1
            dt1 = datetime.datetime.fromtimestamp(time.time())
            print '* Initiating PDA simplification'
            print ' - Total PDA states are ' + repr(len(self.mmc.s))
            handle = IntersectionHandling()
            newpda = handle.get(self.mmc.s, self.mmc.accepted)
            newpda = self.mmc.s
            simply = SimplifyStateIDs()
            newpda, biggestid, newaccepted = simply.get(
                newpda, self.mmc.accepted)
            print ' - Total PDA states after id clearence are ' + repr(len(newpda))
            replace = ReadReplace(newpda, biggestid)
            newpda = replace.replace_read()
            print ' - Total PDA states after read elimination are ' + repr(len(newpda))
            maxstate = replace.nextstate() - 1
            print '* Reduce PDA using DFA BFS (remove unreachable states):'
            reduce_b = ReducePDA()
            newpda = reduce_b.get(newpda)
            print "- Total PDA states after reduction are " + repr(len(newpda))

            dt2 = datetime.datetime.fromtimestamp(time.time())
            rdelta = dateutil.relativedelta.relativedelta(dt2, dt1)
            print "* PDA was simplyfied in %d days, %d hours, %d minutes and %d seconds" % (
                rdelta.days, rdelta.hours, rdelta.minutes, rdelta.seconds)
            dt1 = datetime.datetime.fromtimestamp(time.time())
            print '* Initiating CNF from PDA generation'
            cnfgenerator = PdaCnf(newpda, newaccepted)
            dt2 = datetime.datetime.fromtimestamp(time.time())
            rdelta = dateutil.relativedelta.relativedelta(dt2, dt1)
            print "* CNF was generated in %d days, %d hours, %d minutes and %d seconds" % (
                rdelta.days, rdelta.hours, rdelta.minutes, rdelta.seconds)
            dt1 = datetime.datetime.fromtimestamp(time.time())
            print '* Initiating string from CFG generation'
            grammar = cnfgenerator.get_rules(optimized)
            print ' - Total grammar rules are ' + repr(len(grammar))
            gen = CFGGenerator(cfg.CNF(cfg.gRules(grammar, ":")),
                               optimized=optimized,
                               splitstring=0,
                               maxstate=maxstate)
            return_string = gen.generate()
            dt2 = datetime.datetime.fromtimestamp(time.time())
            rdelta = dateutil.relativedelta.relativedelta(dt2, dt1)
            print "* A string was generated in %d days, %d hours, %d minutes and %d seconds" % (
                rdelta.days, rdelta.hours, rdelta.minutes, rdelta.seconds)

            print return_string
        else:
            return_string = None
        return return_string


def main():
    """
    Testing function for PDA - DFA Diff Operation
    """
    if len(argv) < 2:
        print 'Usage: '
        print '         Get A String              %s CFG_fileA FST_fileB' % argv[0]
        return

    alphabet = createalphabet()

    cfgtopda = CfgPDA(alphabet)
    print '* Parsing Grammar:',
    mma = cfgtopda.yyparse(argv[1])
    print 'OK'

    flex_a = Flexparser(alphabet)
    print '* Parsing Regex:',
    mmb = flex_a.yyparse(argv[2])
    print mmb
    print 'OK'
    print '* Minimize Automaton:',
    mmb.minimize()
    print 'OK'
    print mmb
    print '* Diff:',
    ops = PdaDiff(mma, mmb, alphabet)
    mmc = ops.diff()
    print 'OK'
    print '* Get String:',
    print ops.get_string()


if __name__ == '__main__':
    main()



"""This module contains the DFA implementation"""
# !/usr/bin/python
import imp
import copy
from alphabet import createalphabet
from operator import attrgetter

def bfs(graph, start):
    """
    Finds the shortest string using BFS
    Args:
        graph (DFA): The DFA states
        start (DFA state): The DFA initial state
    Returns:
        str: The shortest string
    """
    # maintain a queue of paths
    queue = []
    visited = []
    # maintain a queue of nodes
    # push the first path into the queue
    queue.append([['', start]])
    while queue:
        # get the first path from the queue
        path = queue.pop(0)
        # get the last node from the path
        node = path[-1][1]
        if node.stateid not in visited:
            visited.append(node.stateid)
            # path found
            if node.final != TropicalWeight(float('inf')):
                return "".join([mnode[0] for mnode in path])
            # enumerate all adjacent nodes, construct a new path and push
            # it into the queue
            for arc in node.arcs:
                char = graph.isyms.find(arc.ilabel)
                next_state = graph[arc.nextstate]
                # print next_state.stateid
                if next_state.stateid not in visited:
                    new_path = list(path)
                    new_path.append([char, next_state])
                    queue.append(new_path)


try:
    print 'Checking for pywrapfst module:',
    imp.find_module('pywrapfst')
    print 'OK'
    from pywrapfstdfa import PywrapfstDFA, TropicalWeight


    class DFA(PywrapfstDFA):
        """The DFA class implemented using openFst library"""

        def __init__(self, alphabet=createalphabet()):
            self.alphabet = alphabet
            super(DFA, self).__init__(alphabet)

        def copy(self):
            mma = DFA(self.alphabet)
            mma.automaton = self.automaton.copy()
            mma.alphabet = copy.deepcopy(self.alphabet)
            mma.isyms = copy.deepcopy(self.isyms)
            mma.osyms = copy.deepcopy(self.osyms)
            return mma

        def shortest_string(self):
            """
            Uses BFS in order to find the shortest string
            Args:
                None
            Returns:
                str: The shortest string
            """
            initialstates = sorted(
                self.states,
                key=attrgetter('initial'),
                reverse=True)
            if len(initialstates) > 0:
                return bfs(self, initialstates[0])
            else:
                return None

        def diff(self, input_mm):
            """
            Automata Diff operation
            """
            mma = DFA(self.alphabet)
            mma.init_from_acceptor(self)
            mmb = DFA(self.alphabet)
            mmb.init_from_acceptor(input_mm)
            mma.minimize()
            mmb.complement(self.alphabet)
            mmb.minimize()
            mmc = DFA(self.alphabet)
            mmc.init_from_acceptor(mma & mmb)
            mmc.minimize()
            return mmc

        def to_regex(self):
            """
            Returns a regex approximation
            Args:
                None
            Returns:
                str: A regex approximation
            """
            from regex import Regex
            converter = Regex(self)
            return converter.get_regex()

except ImportError:
    print 'FAIL'
    try:
        print 'Checking for fst module:',
        imp.find_module('fst')
        print 'OK'
        from fstdfa import FstDFA, TropicalWeight


        class DFA(FstDFA):
            """The DFA class implemented using openFst library"""
            def __init__(self, alphabet = createalphabet()):
                self.alphabet = alphabet
                super(DFA, self).__init__(alphabet)

            def shortest_string(self):
                """
                Uses BFS in order to find the shortest string
                Args:
                    None
                Returns:
                    str: The shortest string
                """
                initialstates = sorted(
                    self.states,
                    key=attrgetter('initial'),
                    reverse=True)
                if len(initialstates) > 0:
                    return bfs(self, initialstates[0])
                else:
                    return None

            def diff(self, input_mm):
                """
                Automata Diff operation
                """
                mma = DFA(self.alphabet)
                mma.init_from_acceptor(self)
                mmb = DFA(self.alphabet)
                mmb.init_from_acceptor(input_mm)
                mma.minimize()
                mmb.complement(self.alphabet)
                mmb.minimize()
                mmc = DFA(self.alphabet)
                mmc.init_from_acceptor(mma & mmb)
                mmc.minimize()
                return mmc

            def to_regex(self):
                """
                Returns a regex approximation
                Args:
                    None
                Returns:
                    str: A regex approximation
                """
                from regex import Regex
                converter = Regex(self)
                return converter.get_regex()

    except ImportError:
        print 'FAIL'
        print 'Fallback to python implementation:OK'
        from pythondfa import PythonDFA, TropicalWeight

        class DFA(PythonDFA):
            """The DFA class implemented using python"""

            def __init__(self, alphabet = createalphabet()):
                self.alphabet = alphabet
                super(DFA, self).__init__(alphabet)

            def copy(self):
                mma = DFA(self.alphabet)
                mma.states = copy.deepcopy(self.states)
                mma.alphabet = copy.deepcopy(self.alphabet)
                mma.isyms = copy.deepcopy(self.isyms)
                mma.osyms = copy.deepcopy(self.osyms)
                return mma

            def shortest_string(self):
                """
                Uses BFS in order to find the shortest string
                Args:
                    None
                Returns:
                    str: The shortest string
                """
                initialstates = sorted(
                    self.states,
                    key=attrgetter('initial'),
                    reverse=True)
                if len(initialstates) > 0:
                    return bfs(self, initialstates[0])
                else:
                    return None

            def diff(self, input_mm):
                """
                Automata Diff operation
                """
                mma = DFA(self.alphabet)
                mma.init_from_acceptor(self)
                mmb = DFA(self.alphabet)
                mmb.init_from_acceptor(input_mm)
                mma.minimize()
                mmb.complement(self.alphabet)
                mmb.minimize()
                mmc = DFA(self.alphabet)
                mmc.init_from_acceptor(mma & mmb)
                mmc.minimize()
                return mmc

            def to_regex(self):
                """
                Returns a regex approximation
                Args:
                    None
                Returns:
                    str: A regex approximation
                """
                from regex import Regex
                converter = Regex(self)
                return converter.get_regex()
